unit TurboReport;

interface

uses
  Printers, Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Buttons, ComCtrls, DBChart, Chart, DBctrls, Extctrls, StdCtrls, DB, DDEMan,
  JPEG, Grids, DBGrids, Dialogs;

type
  TTurboBand       = class;
  TTurboGridBand   = class;
  TTurboReportView = class;

  TGenerateEvent= procedure(Sender: TObject; fast: Boolean) of Object;
  TOnGenerateEvent= procedure(Sender, Item: TObject; fast: Boolean) of Object;
  TReportEvent = procedure(Sender: TObject; NewPage: Integer) of Object;
  TZoomEvent   = procedure(Sender: TObject; Zoom: Integer) of Object;

  TTurboPageSize  = (A3, A4, A5, B4, B5); // Kat Boyutu
  TTurboPrintMode = (prNormal, prFast);   // Yazm Modu
  TTurboBandType  = (rbDetail, rbSubDetail, rbSummary,
                      rbHeaderFooter, rbPageHeader, rbPageFooter); // Band Tipi

  TPrintCanvas = TMetafileCanvas;
  TPrintPage   = TMetafile;
  TNavigatorDirection = (dirAuto, dirHorizontal, dirVertical);
  TPageCtlButton = (pgFirst, pgPrior, pgNext, pgLast, pgZoomIn, pgZoomOut,
                    pgPrint, pgSetupPrinter, pgExcel, pgHTML, pgReFresh);
  EPBtnClick = procedure (Sender: TObject; Button: TPageCtlButton) of object;
  TPageButtonSet = set of TPageCtlButton;

  TSysData     = (sdDate, sdDateTime, sdTime, sdCurrentPage);

  TTurboExcel = class(TComponent)
  private
      FMacro     : TFileName;
      FMacroPath : TFileName;
      FDDE       : TDdeClientConv;
      FConnected : Boolean;
      FExeName   : TFileName;
      FDecimals  : Integer;
      FOnClose   : TNotifyEvent;
      FOnOpen    : TNotifyEvent;
      FBatch     : Boolean;
      FMin       : Integer;
      FMax       : Integer;
      FFirstRow  : Integer;
      FFirstCol  : Integer;
      FLastCol   : Integer;
      FLines     : TStrings;    { using TStringList }
      FCells     : TStrings;    { using TStringList }

    procedure SetExeName(const Value: TFileName);
    procedure SetConnect(const Value: Boolean);
    procedure SetMin    (const Value: Integer);
    procedure SetMax    (const Value: Integer);
    function GetSelection: string;
    function GetReady: Boolean;
  protected
    procedure DoRect(Top, Left, Bottom, Right: Integer;
                     Data: TStrings; Request: Boolean);
    procedure CheckConnection; virtual;
    procedure LinkSystem;
    procedure OpenLink(Sender: TObject);
    procedure ShutDown(Sender: TObject);
    procedure LocateExcel; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Connect;
    procedure Disconnect;
    procedure Wait;
    procedure ProcessMessages;
    function Request(const Item: string): string;
    procedure Exec  (const Cmd : string);
    procedure Run   (const Mn  : string);
    procedure Select(Row, Col: Integer);
    procedure PutStr(Row, Col: Integer; const s: string);
    procedure PutExt(Row, Col: Integer; e: Extended); virtual;
    procedure PutMoney(Row, Col: Integer; e: Double); virtual;
    procedure PutInt(Row, Col: Integer; i: Longint);  virtual;
    procedure PutDay(Row, Col: Integer; d: TDateTime);virtual;
    procedure BatchStart(FirstRow, FirstCol: Integer);
    procedure BatchCancel;
    procedure BatchSend;
    procedure GetRange(R: TRect; Lines: TStrings);
    function GetCell(Row, Col: Integer): string;
    procedure OpenMacroFile(const Fn: TFileName; Hide: Boolean);
    procedure CloseMacroFile;
    property DDE: TDdeCLientConv   read FDDE;
    property Connected: Boolean    read FConnected write SetConnect;
    property Ready    : Boolean    read GetReady;
    property Selection: string     read GetSelection;
    property Lines    : TStrings   read FLines;
    property FirstRow : Integer    read FFirstRow;
    property LastCol  : Integer    read FLastCol   write FLastCol;
    property BatchOn  : Boolean    read FBatch;
  published
    property ExeName  : TFileName  read FExeName   write SetExeName;
    property Decimals : Integer    read FDecimals  write FDecimals;
    property BatchMin : Integer    read FMin       write SetMin;
    property BatchMax : Integer    read FMax       write SetMax;
    property OnClose: TNotifyEvent read FOnClose   write FOnClose;
    property OnOpen : TNotifyEvent read FOnOpen    write FOnOpen;
  end;

  TTurboPageList    = class(TList)
  public
    destructor Destroy; override;
    function GetPage(const Index: Integer): TPrintPage;
  end;

  TTurboMargins = Class(TPersistent)
  private
    fLeft,
    fRight,
    fBottom,
    fTop,
    fColumn,
    fColumnSpace: Integer;
    fSizeChange: TNotifyEvent;

    Procedure setColumn(Value: Integer);
    Procedure SetColumnSpace(Value: Integer);
    Procedure SetLMargin(Value: Integer);
    Procedure SetRMargin(Value: Integer);
    Procedure SetBMargin(Value: Integer);
    Procedure SetTMargin(Value: Integer);
    Procedure DataChange;
  public
    Constructor Create;
  published
    property Left: Integer Read fLeft Write SetLMargin;      // Sol Boluk (Sadece prNormal)
    property Right: Integer Read fRight Write SetRMargin;    // Sa Boluk (Sadece prNormal)
    property Top: Integer Read fTop Write SetTMargin;        // st Boluk (Sadece prNormal)
    property Bottom: Integer Read fBottom Write SetBMargin;  // Alt Boluk (Sadece prNormal)
    property Column: Integer Read fColumn Write SetColumn;   // Sayfadaki Kolon Says
    property ColumnSpace: Integer Read fColumnSpace Write SetColumnSpace; // Kolonlar Aras Boluk (Pixel)
    property OnSizeChange: TNotifyEvent read fSizeChange write fSizeChange; // Boluklar Deiirse;
  End;

  TTurboPrinter = Class(TPersistent)
  private
    fColWidth,
    fColHeight,
    fWidth,
    fHeight: Integer;
    fPort,
    fTitle: String;
    fOrientation: TPrinterOrientation;
    fPaperSize  : TTurboPageSize;
    fSizeChange : TNotifyEvent;
    fMode       : TTurboPrintMode;

    Procedure SetOrientation(Value: TPrinterOrientation);
    Procedure setPaperSize(Value: TTurboPageSize);
  public
    Constructor Create;
  published
    property Width: Integer read fWidth;       // Genilik (Pixel)
    property Height: Integer read fHeight;    // Ykseklik (Pixel)
    property ColWidth: Integer read fColwidth write fColwidth;       // Kolon Says (Sadece prFast)
    property ColHeight: Integer read fColHeight write fColHeight;    // Satr Says (Sadece prFast)
    property Port: String read fPort write fPort;                  // Yazc Portu (Sadece prFast)
    property Mode: TTurboPrintMode read fMode write fMode; // Yazm Modu
    property Title: String read fTitle write fTitle;        // Windows in Balk
    property Orientation: TPrinterOrientation read fOrientation write setOrientation; // Yazdrma Yn
    property PaperSize  : TTurboPageSize read fPaperSize write setPaperSize; // Kat Boyutu (Sadece prNormal)
    property OnSizeChange: TNotifyEvent read fSizeChange write fSizeChange; // Boluklar Deiirse;
  End;

  TTurboFrame = Class(TPersistent)
  private
    fBottom,
    fTop,
    fLEft,
    fRight: Boolean;
    fPen: TPen;
    fFrameChange: TNotifyEvent;

    Procedure SetBottom(Value: Boolean);
    Procedure SetLeft(Value: Boolean);
    Procedure SetRight(Value: Boolean);
    Procedure SetTop(Value: Boolean);
    Procedure PenChange(Sender: TObject);
  Public
    constructor Create;
    Destructor Free;
  published
    property Bottom: Boolean Read fBottom Write setBottom default False;
    property Top   : Boolean Read fTop Write setTop default False;
    property Left  : Boolean Read fLeft Write setLeft default False;
    property Right : Boolean read fRight write setRight default False;
    Property Pen   : TPen Read FPen Write FPen;
    property OnFrameChange: TNotifyEvent Read fFrameChange write fFrameChange;
  End;

  TTurboReport = class(TCustomControl)
  private
    fPrinter   : TTurboPrinter;
    fMargins   : TTurboMargins;
    fDataSource: TDataSource;
    fSizeChange: TNotifyEvent;

    Procedure OnMarginsChange(Sender: TObject);
    Procedure OnPageSizeChange(Sender: TObject);
    { Private declarations }
  protected
    procedure Paint; override;
    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
    { Protected declarations }
  public
    Constructor Create(AOwner: TComponent); override;
    Destructor Free;
    { Public declarations }
  published
    Property Printer: TTurboPrinter Read fPrinter write fPrinter;
    Property Margins: TTurboMargins Read fMargins Write fMargins;
    Property OnSizeChange: TNotifyEvent Read FSizeChange Write fSizeChange;
    property DataSource: TDataSource Read fDataSource write fDataSource;
    { Published declarations }
  end;

  TTurboSysData = Class(TCustomLabel)
  private
    fReportViewer: TTurboReportView;
    fSysData   : TSysData;
    Function getCaption: String;
  protected
    procedure Paint; Override;
  public
    constructor Create(AOwner: TComponent); Override;
  published
    property Alignment;
    property Caption;
    property Color;
    property Font;
    property SysData: TSysData Read fSysData write fSysData;
    property ReportViewer: TTurboReportView read fReportViewer write fReportViewer;
  End;

  TTurboBand = Class(TCustomControl)
  private
    fBandType  : TTurboBandType;
    fDataSource: TDataSource;
    fFooterBand,
    fHeaderBand: TTurboBand;
    fFrame     : TTurboFrame;
    fBeforeGenerate,
    fAfterGenerate: TGenerateEvent;
    fOnGenerate: TOnGenerateEvent;
    fBrush: TBrush;

    procedure setBandType(Value: TTurboBandType);
    Procedure FrameChange(Sender: TObject);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    Procedure GetTop(O: TTurboReport);
    Procedure GetWidth(O: TTurboReport);
    procedure WMMove(var Message: TWMMove); message WM_MOVE;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
    procedure Paint; Override;
  public
    Constructor Create(AOwner: TComponent); override;
    Destructor Free;
    Procedure GetSizes;
  published
    property Brush: TBrush read fBrush write fBrush;
    property HeaderBand: TTurboBand read fHeaderBand Write fHeaderBand;
    property FooterBand: TTurboBand read fFooterBand Write fFooterBand;
    property Frame: TTurboFrame  Read fFrame write fFrame;
    property BandType: TTurboBandType Read fBandtype Write setBandType;
    property DataSource: TDataSource Read fDataSource write fDataSource;
    property AfterGenerate:TGenerateEvent Read fAfterGenerate write fAfterGenerate;
    property BeforeGenerate:TGenerateEvent Read fbeforeGenerate write fBeforeGenerate;
    property OnGenerate: TOnGenerateEvent Read fOnGenerate write fOnGenerate;
  End;

  TTurboGridBand = class(TTurboBand)
    fGrid: TCustomGrid;

    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  published
    property  Grid: TCustomGrid Read fGrid write fGrid;
  end;

  TTurboReportGenerator = Class(TObject)
    protected
      fReport: TTurboReport;

      Procedure DrawLabel(Band: TTurboBand; L: TCustomLabel); virtual;
      Procedure DrawImage(Band: TTurboBand; Img: TImage); virtual;
      Procedure DrawDBImage(Band: TTurboBand; Img: TDBImage); virtual;
      Procedure DrawShape(Band: TTurboBand; Sp: TShape); virtual;
      Procedure DrawMemo(Band: TTurboBand; M: TCustomMemo); virtual;
      Procedure DrawChart(Band: TTurboBand; M: TChart); virtual;
      Procedure DrawSysData(Band: TTurboBand; D: TTurboSysdata); virtual;

      procedure GenerateGridBand(Band: TTurboGridBand); virtual;
      procedure GenerateBand(Band: TTurboBand); virtual;
      procedure GenerateBandData(Band: TTurboBand); virtual;
      Procedure Control(Band: TTurboBand); Virtual;
      Procedure Calculate(Band: TTurboBand); Virtual;
      Procedure DrawFrame(Band: TTurboBand); Virtual;
      Procedure GenerateBandforType(BandType: TTurboBandType); Virtual;
      procedure newPage; Virtual;
      procedure BeforeGenerate; Virtual;
      procedure AfterGenerate; Virtual;
      procedure BeforeGenerateData(Band: TTurboBand); Virtual;
      procedure AfterGenerateData(Band: TTurboBand); Virtual;
    public
      procedure Generate(Report: TTurboReport); virtual;
  end;

  TTurboWMFGenerator = Class (TTurboReportGenerator)
  Protected
    CurrentCol,             // Current Column
    HHeader,                // Header Width
    HFooter: Integer;       // Footer Width
    PageList   : TTurboPageList;
    fCurrentPage,
    XStart,
    YStart: Integer;
    ReportPage: TPrintPage;

    function  GetPages(Indx: Cardinal): TPrintPage;
    function  GetPageCount: Integer;
    Function  RectControl(Band: TTurboBAnd; Var R: TRect; Var B: Boolean): TRect;
    Procedure ColumnControl(Band: TTurboBand);
    Procedure Calculate(Band: TTurboBand); override;
    procedure newPage; override;
    procedure BeforeGenerate; override;
    procedure AfterGenerate; override;
    Procedure DrawFrame(Band: TTurboBand); Override;

    Procedure Control(Band: TTurboBand); override;
    Procedure DrawLabel(Band: TTurboBand; L: TCustomLabel); override;
    Procedure DrawImage(Band: TTurboBand; Img: TImage); override;
    Procedure DrawDBImage(Band: TTurboBand; Img: TDBImage); override;
    Procedure DrawMemo(Band: TTurboBand; M: TCustomMemo); override;
    Procedure DrawShape(Band: TTurboBand; Sp: TShape); Override;
    Procedure DrawChart(Band: TTurboBand; M: TChart); Override;
    procedure BeforeGenerateData(Band: TTurboBand); override;
  public
    ReportCanvas: TPrintCanvas;  // Canvas For WMF File
    property Pages[Indx: Cardinal]: TPrintPage read GetPages;
    Property PageCount : Integer Read GetPageCount;

    Constructor Create;
    Destructor Destroy; override;
  End;

  TTurboTextGenerator = Class (TTurboReportGenerator)
  Protected
    CurrentCol,             // Current Column
    HHeader,                // Header Genilii
    HFooter: Integer;       // Footer Genilii
    PageList   : TStringList;
    fCurrentPage,
    XStart,
    YStart: Integer;

    procedure newPage; override;
    Procedure DrawLabel(Band: TTurboBand; L: TCustomLabel); override;
    Procedure DrawMemo(Band: TTurboBand; M: TCustomMemo); override;
    procedure BeforeGenerate; override;
    Procedure ColumnControl(Band: TTurboBand);
    Procedure Calculate(Band: TTurboBand); override;
    Function getCurLine: Integer;
  public
    Property Lines: TStringList Read PageList;

    Constructor Create;
    Destructor Destroy; override;
  End;

  TTurboExcelGenerator = Class (TTurboReportGenerator)
  Protected
    PageList   : TStringList;
    XStart,
    YStart: Integer;

    Procedure DrawLabel(Band: TTurboBand; L: TCustomLabel); override;
    Procedure DrawMemo(Band: TTurboBand; M: TCustomMemo); override;
    procedure BeforeGenerateData(Band: TTurboBand); override;
    Procedure Calculate(Band: TTurboBand); override;
    Function getCurLine: Integer;
  public
    Property Lines: TStringList Read PageList;

    Constructor Create;
    Destructor Destroy; override;
    Procedure Exec;
  End;

  TTurboHTMLGenerator = Class (TTurboReportGenerator)
  private
    PageList: TStringList;
    picCount: Integer;
    fName,
    fPrnTitle: String;

    procedure BeforeGenerateData(Band: TTurboBand); override;
    procedure AfterGenerateData(Band: TTurboBand); override;
    procedure BeforeGenerate; override;
    procedure AfterGenerate; override;
    Function getFontStr(F: TFont): String;
    Function getFontEndStr(F: TFont): String;
    Function ControlColor(C: TColor): String;
    Function ControlProperties(C: TControl): String;
    Procedure SaveImage(Img: TGraphic; ImgClass: TGraphicClass);

    Procedure DrawLabel(Band: TTurboBand; L: TCustomLabel); override;
    Procedure DrawImage(Band: TTurboBand; Img: TImage); override;
    Procedure DrawDBImage(Band: TTurboBand; Img: TDBImage); override;
    Procedure DrawMemo(Band: TTurboBand; M: TCustomMemo); override;
//    Procedure DrawShape(Band: TTurboBand; Sp: TShape); Override;
    Procedure DrawChart(Band: TTurboBand; M: TChart); Override;
  public
    Property Lines: TStringList Read PageList;
    property fileName    : String Read fName write fName;
    property PrinterTitle: String Read fPrnTitle write fPrnTitle;

    Constructor Create;
    Destructor Destroy; override;
  End;

  TTurboReportView = class(TScrollBox)
  private
    fReportName,
    fReportDirectory: String;
    fZoom,
    fCurrentPage,
    fPageCount  : Integer;
    fReport     : TTurboReport;
    fPaintBox   : TPaintBox;
    fProgressBar: TProgressBar;
    fPageChange : TReportEvent;
    fAfterPrint : TReportEvent;
    fBeforePrint: TReportEvent;
    fPrint      : TReportEvent;
    fZoomChange : TZoomevent;
    fReportGenerator: TTurboWMFGenerator;

    Procedure SetCurrentPage(Value: Integer);
    Procedure SetZoom(Value: Integer);
    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
  protected
    procedure PaintPreview(Sender: TObject); virtual;
  public
    property CurrentPage: Integer Read fCurrentPage Write setCurrentPage;
    constructor Create(AOwner: TComponent);  Override;
    Destructor Free;

    Procedure Generate;
    Procedure GenerateText(L: TStrings; fSave: Boolean);
    Procedure GenerateHTML(L: TStrings; fSave: Boolean);
    Procedure GenerateExcel(L: TStrings);
    Procedure SetupPrinter;
    Procedure Print;

    Procedure Next;
    Procedure First;
    Procedure Last;
    Procedure Prev;

  published
    Property ReportName: String read fReportName write fReportName;
    Property ReportDirectory: String read fReportDirectory write fReportDirectory;
    Property ProgressBar: TProgressBar read fProgressBar write fProgressBar;
    property Report: TTurboReport Read fReport Write fReport;
    property PageCount: Integer Read fPageCount write fPageCount;
    Property Zoom: Integer Read fZoom Write setZoom;

    Property OnPageChange : TReportEvent read fPageChange write fPageChange;
    Property OnAfterPrint : TReportEvent read fAfterPrint write fAfterPrint;
    Property OnBeforePrint: TReportEvent read fBeforePrint write fBeforePrint;
    Property OnPrint      : TReportEvent read fPrint write fPrint;
    Property OnZoomChange : TZoomevent   read fZoomChange write fZoomChange;
  End;

  TTurboBtn = class(TSpeedButton)
  private
    FIndex: TPageCtlButton;
  published
    property Index: TPageCtlButton read fIndex write FIndex;
  End;

  TTurboNavigator = class (TCustomPanel)
  private
    FOnBtnClick: EPBtnClick;
    fReportViewer: TTurboReportView;
    FHints: TStrings;
    ButtonWidth: Integer;
    MinBtnSize: TPoint;
    FFlat: Boolean;
    FVisibleButtons: TPageButtonSet;
    FocusedButton: TPageCtlButton;
    FDirection : TNavigatorDirection;

    procedure ClickHandler(Sender: TObject);
    procedure InitButtons;
    procedure InitHints;
     procedure BtnMouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetVisible(Value: TPageButtonSet);
    procedure AdjustSize (var W: Integer; var H: Integer);
    procedure HintsChanged(Sender: TObject);
    procedure SetHints(Value: TStrings);
    procedure SetFlat(Value: Boolean);
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  protected
    Buttons: array[TPageCtlButton] of TTurboBtn;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure ActiveChanged;
    procedure Loaded; override;
    procedure FSetDirection(Value: TNavigatorDirection);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure BtnClick(Index: TPageCtlButton);
  published
    property Direction : TNavigatorDirection read fDirection write FSetDirection default dirAuto;
    property VisibleButtons: TPageButtonSet read FVisibleButtons write SetVisible
      default [pgFirst, pgPrior, pgNext, pgLast, pgZoomIn, pgZoomOut, pgPrint,
                 pgSetupPrinter, pgExcel, pgHTML, pgReFresh];
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Flat: Boolean read FFlat write SetFlat default False;
    property Ctl3D;
    property Hints: TStrings read FHints write SetHints;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnResize;
    property OnStartDrag;
    property OnClick: EPBtnClick read FOnBtnClick write FOnBtnClick;
    property ReportViewer: TTurboReportView read fReportViewer write fReportViewer;
  end;

{$R RptCtrls.Res}

procedure Register;
Procedure AddFieldFromGrid(Band: TTurboBand; Grid: TDBGrid);
implementation
Uses dbtables;
Const
  BuffSize         = 512;
  msgNoExcel       = 'Excel Bulunamyor...';
  msgNoConnect     = 'Excel baplants yaplamad...';
  msgNoLink        = 'Excel balants geersiz...';
  msgNoRespond     = 'Excel not responding';
  msgNoReply       = '*** No Reply ***';
  msgNotAccepted   ='" not accepted by Excel';
  msgNoMacro       = 'Macro alamad';
  msgNoTable       = 'Dosya alamad';

Var
  CharYInPixel, CharWInPixel, PixelsInMM, PixelsInInch: Single;
  BtnTypeName: array[TPageCtlButton] of PChar = ('FIRST', 'PRIOR', 'NEXT',
    'LAST', 'ZOOMIN', 'ZOOMOUT', 'PRINT', 'SETUP', 'EXCEL', 'HTML', 'REFRESH');
  BtnHintId: array[TPageCtlButton] of String = ('lk Sayfa', 'nceki Sayfa',
     'Sonraki Sayfa', 'Son Sayfa', 'Byt', 'Klt', 'Yazdr', 'Ayarlar',
     'Excel e Transfer', 'HTML olarak kaydet', 'Yenile');



procedure Register;
begin
  RegisterComponents('TurboReport', [TTurboReport, TTurboBand, TTurboGridBand,
    TTurboReportView, TTurboNavigator, TTurboSysData, TTurboExcel]);
end;

destructor TTurboPageList.Destroy;
var
 I: Integer;
 Page: TPrintPage;
begin
  for I:=0 to Count-1 do
  begin
    Page:= GetPage(i);
    if Page <> nil then Page.Free;
  end;
  inherited Destroy;
end;

function TTurboPageList.GetPage(const Index: Integer): TPrintPage;
begin
   Result:= TPrintPage(Items[Index]);
end;

Procedure TTurboFrame.PenChange(Sender: TObject);
Begin
  if Assigned(onFrameChange) then OnFrameChange(Self);
End;

constructor TTurboFrame.Create;
Begin
  Inherited;
  fPen:= TPen.Create;
  fPen.OnChange:= PenChange;
End;

Destructor TTurboFrame.Free;
Begin
  fPen.Free;
  Inherited Free;
End;

Procedure TTurboFrame.SetBottom(Value: Boolean);
Begin
  if Value <> fBottom Then
  Begin
    fBottom:= Value;
    if Assigned(onFrameChange) then OnFrameChange(Self);
  End;
end;

Procedure TTurboFrame.SetLeft(Value: Boolean);
Begin
  if Value <> fLeft Then
  Begin
    fLeft := Value;
    if Assigned(onFrameChange) then OnFrameChange(Self);
  end;
end;

Procedure TTurboFrame.SetRight(Value: Boolean);
Begin
  if Value <> fRight Then
  Begin
    fRight:= Value;
    if Assigned(onFrameChange) then OnFrameChange(Self);
  End;
end;

Procedure TTurboFrame.SetTop(Value: Boolean);
Begin
  if Value <> fTop Then
  Begin
    fTop:= Value;
    if Assigned(onFrameChange) then OnFrameChange(Self);
  End;
end;

Constructor TTurboMargins.Create;
Begin
  inherited;
  fLeft:= 0;
  fRight:= 0;
  fBottom:= 0;
  fTop:= 0;
  fColumn:= 1;
  fColumnSpace:= 0;
End;

Procedure TTurboMargins.setColumn(Value: Integer);
Begin
  if (Value <> fColumn) and (Value > 0) Then
  Begin
    fColumn:= Value;
    DataChange;
  end;
End;

Procedure TTurboMargins.SetColumnSpace(Value: Integer);
Begin
  if Value <> fColumnSpace Then
  Begin
    fColumnSpace:= Value;
    DataChange;
  end;
End;

Procedure TTurboMargins.SetLMargin(Value: Integer);
Begin
  if Value <> fLeft Then
  Begin
    fLeft:= Value;
    DataChange;
  end;
End;

Procedure TTurboMargins.SetRMargin(Value: Integer);
Begin
  if Value <> fRight Then
  Begin
    fRight:= Value;
    DataChange;
  end;
End;

Procedure TTurboMargins.SetBMargin(Value: Integer);
Begin
  if Value <> fBottom Then
  Begin
    fBottom:= Value;
    DataChange;
  end;
End;

Procedure TTurboMargins.SetTMargin(Value: Integer);
Begin
  if Value <> fTop Then
  Begin
    fTop:= Value;
    DataChange;
  end;
End;

Procedure TTurboMargins.DataChange;
Begin
  if Assigned (OnSizeChange) Then OnSizeChange(Self);
End;

Constructor TTurboPrinter.Create;
Begin
  inherited;
  fColWidth:= 80;
  fColHeight:= 55;
  fWidth:= 0;
  fHeight:= 0;
  fPort:= 'LPT1';
  fTitle:= '';
  fOrientation:= poPortrait;
  fPaperSize  := A4;
  fMode:= prNormal;
End;

Procedure TTurboPrinter.SetOrientation(Value: TPrinterOrientation);
Var
  T: Integer;
Begin
  if Value <> fOrientation Then
  Begin
    fOrientation:= Value;
    T:= fHeight;
    fHeight:= fWidth;
    fWidth := T;
    if Assigned (OnSizeChange) Then OnSizeChange(Self);
  End;
End;

Procedure TTurboPrinter.setPaperSize(Value: TTurboPageSize);
Var
  Tmp: Integer;
Begin
  if Value <> fPaperSize Then
  Begin
    Case Value Of
      A3: Begin fHeight:= 420; fWidth:= 297; End;
      A4: Begin fHeight:= 297; fWidth:= 210; End;
      A5: Begin fHeight:= 210; fWidth:= 148; End;
      B4: Begin fHeight:= 354; fWidth:= 250; End;
      B5: Begin fHeight:= 257; fWidth:= 184; End;
    End;
    if Orientation = poLandscape Then
    Begin
      Tmp    := fHeight;
      fHeight:= fWidth;
      fWidth := Tmp;
    End;
    fPaperSize:= Value;
    if Assigned (OnSizeChange) Then OnSizeChange(Self);
  End;
End;

constructor TTurboBand.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  ControlStyle:= ControlStyle + [csAcceptsControls];
  Color:= clNone;
  Height:= 40;
  fFrame:= TTurboFrame.Create;
  fFrame.OnFrameChange:= FrameChange;
  fBandType:= rbDetail;
  fDataSource:= nil;
  fBrush:= TBrush.Create;
  fBrush.OnChange:= FrameChange;
  GetSizes;
End;

Destructor TTurboBand.Free;
Begin
  fBrush.Create;
  fFrame.Free;
  Inherited Free;
End;

procedure TTurboBand.setBandType(Value: TTurboBandType);
Begin
  if Value <> fBandType Then
  Begin
    fBandType:= Value;
    GetSizes;
  End;
End;

procedure TTurboBAnd.WMMove(var Message: TWMMove);
Begin
  GetSizes;
End;

procedure TTurboBAnd.WMSize(var Message: TWMSize);
Begin
  GetSizes;
End;

procedure TTurboBand.Paint;
Procedure DrawMeter;
Var
  N: Integer;
  X: Single;
Begin
  with Canvas do
  Begin
    Pen.Width:= 1;
    Pen.Style:= psSolid;
    X:= PixelsInInch;
    N:= 1;
    while X < Width do
    begin
      MoveTo(Trunc(X), 1);
      LineTo(Trunc(X), Height);
      X := X + PixelsInInch;
      N:= N + 1;
      if (N > 0) and (N Mod 5 = 0 ) Then Pen.Color:= clGreen
      else Pen.Color:= clSilver;
    End;
    X:= PixelsInInch;
    Pen.Color:= clSilver;
    N:= 1;
    while X < Height do
    Begin
      MoveTo(1, Trunc(X));
      LineTo(Width, Trunc(X));
      X := X + PixelsInInch;
      N:= N + 1;
      if (N > 0) and (N Mod 5 = 0 ) Then Pen.Color:= clGreen
      else Pen.Color:= clSilver;
    End;
  End;
End;

Procedure DrawLine(X1, Y1, X2, Y2: Integer);
Begin
  With Canvas Do
  Begin
    MoveTo(X1, Y1);
    LineTo(X2, Y2);
  End;
End;

Procedure DrawFrames;
Var
  W: Integer;
Begin
  W:= Frame.Pen.Width div 2;
  Canvas.Pen.Assign(Frame.Pen);
  if Frame.Left   Then DrawLine(0, 0, 0, Height - W);
  if Frame.Top    Then DrawLine(0, 0, Width - W, 0);
  if Frame.Bottom Then DrawLine(0, Height - W, Width - W, Height - W);
  if Frame.Right  Then DrawLine(Width - W, 0, Width - W, Height - W);
End;

Var
  B: TBrush;
Begin
  B:= TBrush.Create;
  B.Assign(Canvas.Brush);
  Canvas.Brush.Assign(Brush);
  if csDesigning in ComponentState Then
   With Canvas do
   Begin
     Pen.Color:= clBlack;
     Pen.Style:= psDashDotDot;
     Pen.Width:= 1;
     Rectangle(0, 0, Width, Height);
     FillRect(Rect(1, 1, Width - 1, Height - 1));
     DrawMeter;
     if not Frame.Left   Then DrawLine(0, 0, 0, Height);
     if not Frame.Top    Then DrawLine(0, 0, Width, 0);
     if not Frame.Bottom Then DrawLine(0, Height, Width, Height);
     if not Frame.Right  Then DrawLine(Width, 0, Width, Height);
     Pen.Assign(Frame.Pen);
     DrawFrames;
   End;
   Canvas.Brush.Assign(B);
   B.Free;
end;

Procedure TTurboBand.GetTop(O: TTurboReport);
Var
  T, I: Integer;
Begin
  T:= Trunc(O.Margins.Top * PixelsInMM);
  For I:= 0 to O.ControlCount - 1 Do
    if (O.Controls[I] is TTurboBand) Then
    Begin
      TTurboBand(O.Controls[I]).Top:= T;
      T:= T + TTurboBand(O.Controls[I]).Height;
    End;
End;

Procedure TTurboBand.GetWidth(O: TTurboReport);
Var
  W, I: Integer;
Begin
  W:= 0;
  With O.Margins Do
  Begin
    I:= Trunc(((ColumnSpace * (Column - 1)) + Left + Right) * PixelsInMM);
    I:= (O.Width - I) div Column;
    W:= I;
  End;
  Width:= W;
End;

Procedure TTurboBand.GetSizes;
Var
  C: TTurboReport;
Begin
  C:= nil;
  if Parent is TTurboReport Then
  Begin
    C:= TTurboReport(Parent);
    Left:= Trunc(C.Margins.Left * PixelsInMM);
    if (BandType = rbPageHeader) or (BandType = rbPageFooter) Then
       Width:= C.Width -
               Trunc(C.Margins.Right * PixelsInMM) -
               Trunc(C.Margins.Left * PixelsInMM)
    else getWidth(C);
    getTop(C);
  End
  else Width:= 150;
End;

Procedure TTurboBand.FrameChange(Sender: TObject);
Begin
  Invalidate;
end;

procedure TTurboBand.Notification(AComponent: TComponent;
      Operation: TOperation);
Begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) Then
    if (AComponent = DataSource) then DataSource := nil
    else if (AComponent = HeaderBand) Then HeaderBand:= nil
    else if (AComponent = FooterBand) Then FooterBand:= nil;
End;

Constructor TTurboReport.Create(AOwner: TComponent);
Begin
  Inherited;
  fPrinter:= TTurboPrinter.Create;
  fPrinter.OnSizeChange:= OnPageSizeChange;

  fMargins:= TTurboMargins.Create;
  fMargins.OnSizeChange:= OnMarginsChange;

  ControlStyle := ControlStyle + [csAcceptsControls];
  Color:= clWhite;
  Height:= Trunc(297 * 3.7762);
  Width := Trunc(210 * 3.7762);
End;

Destructor TTurboReport.Free;
Begin
  fPrinter.Free;
  fMargins.Free;
  Inherited Free;
End;

Procedure TTurboReport.OnMarginsChange(Sender: TObject);
Var
  I: Integer;
Begin
  Invalidate;
  For I:= 0 to ControlCount - 1 Do
    if Controls[I] is TTurboBand Then
    Begin
      TTurboBAnd(Controls[I]).GetSizes;
      TTurboBAnd(Controls[I]).Invalidate;
      Break;
    End;

  For I:= 0 to ControlCount - 1 Do
    if Controls[I] is TTurboBand Then
       TTurboBAnd(Controls[I]).Invalidate;
  if Assigned (OnSizeChange) then OnSizeChange(Self);
End;

Procedure TTurboReport.OnPageSizeChange(Sender: TObject);
Var
  I: Integer;
Begin
  Height:= Trunc(fPrinter.Height * 3.7762);
  Width := Trunc(fPrinter.Width  * 3.7762);
  Invalidate;
  For I:= 0 to ControlCount - 1 Do
    if Controls[I] is TTurboBand Then
    Begin
      TTurboBAnd(Controls[I]).GetSizes;
      TTurboBAnd(Controls[I]).Invalidate;
      Break;
    End;
  For I:= 0 to ControlCount - 1 Do
    if Controls[I] is TTurboBand Then
      TTurboBAnd(Controls[I]).Invalidate;
  if Assigned (OnSizeChange) then OnSizeChange(Self);
End;

procedure TTurboReport.Paint;
Var
  X: Single;
  N: Word;
Begin
  with Canvas do
  begin
    Pen.Color  := clBlack;
    Brush.Color:= Color;
    FillRect(ClientRect);
    with ClientRect do Rectangle(Left, Top, Right, Bottom);
    Pen.Color  := clSilver;
    Font.Size  := 4;
    Font.Name  := 'MS Serif';
    Font.Color:= clBlue;
    X:= PixelsInInch;
    N:= 1;
    while X < Width do
    begin
      MoveTo(Trunc(X), 1);
      if (N > 0) and (N mod 5 = 0) then
      Begin
        Pen.Color:= clGreen;
        LineTo(Trunc(X), Height);
        Pen.Color:= clSilver;
        TextOut(PenPos.X + 1, 2, IntToStr(N div 5));
      End
      Else LineTo(Trunc(X), Height);
      N := N + 1;
      X := X + PixelsInInch;
    End;
    N:= 1;
    X:= PixelsInInch;
    while X < Height do
    Begin
      MoveTo(1, Trunc(X));
      if (N > 0) and (N mod 5 = 0) then
      Begin
        Pen.Color:= clGreen;
        LineTo(Width, Trunc(X));
        Pen.Color:= clSilver;
        TextOut(1, PenPos.Y + 1, IntToStr(N div 5));
      End
      Else LineTo(Width, Trunc(X));
      N := N + 1;
      X := X + PixelsInInch;
    End;
    Pen.Color:= ClBlack;
    Pen.Style:= psDashDotDot;
    Brush.Style:= bsClear;
    Rectangle(Trunc(Margins.Left * PixelsInMM),
              Trunc(Margins.Top * PixelsInMM),
              Width - Trunc(Margins.Right * PixelsInMM),
              Height - Trunc(Margins.Bottom * PixelsInMM));
    Pen.Style:= psSolid;
  End;
End;

procedure TTurboReport.Notification(AComponent: TComponent; Operation: TOperation);
Var
  I: Integer;
Begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and
     (AComponent = DataSource) then DataSource := nil
  else
    For I:= 0 to ControlCount - 1 Do
      if Controls[I] is TTurboBand Then
      Begin
        TTurboBAnd(Controls[I]).GetSizes;
        Break;
      End;
End;

Procedure TTurboReportGenerator.DrawFrame(Band: TTurboBand);
Begin
End;

procedure TTurboReportGenerator.GenerateBandData(Band: TTurboBand);
Var
  I: Integer;
Begin
  For I:= 0 to Band.ControlCount - 1 Do
  Begin
    if Assigned(Band.OnGenerate) Then Band.OnGenerate(Band, Band.Controls[I], False);
    if Band.Controls[I] is TLabel Then
       DrawLabel(Band, TLabel(Band.Controls[I]))
    else if Band.Controls[I] is TDBText Then
         DrawLabel(Band, TDBText(Band.Controls[I]))
    else if Band.Controls[I] is TDBMemo Then
         DrawMemo(Band, TCustomMemo(Band.Controls[I]))
    else if Band.Controls[I] is TShape Then
         DrawShape(Band, TShape(Band.Controls[I]))
    else if Band.Controls[I] is TImage Then
         DrawImage(Band, TImage(Band.Controls[I]))
    else if Band.Controls[I] is TMemo Then
         DrawMemo(Band, TMemo(Band.Controls[I]))
    else if Band.Controls[I] is TDBImage Then
         DrawDBImage(Band, TDBImage(Band.Controls[I]))
    else if Band.Controls[I] is TChart Then
         DrawChart(Band, TChart(Band.Controls[I]))
    else if Band.Controls[I] is TDBChart Then
         DrawChart(Band, TChart(Band.Controls[I]))
    else if Band.Controls[I] is TTurboSysData Then
         DrawSysData(Band, TTurboSysData(Band.Controls[I]))
  End;
End;

Procedure TTurboReportGenerator.DrawSysData(Band: TTurboBand; D: TTurboSysdata);
Begin
  case D.SysData of
    sdDate       : D.Caption:= DateToStr(Date);
    sdDateTime   : D.Caption:= DateToStr(Date) + ' , ' + TimeToStr(Time);
    sdTime       : D.Caption:= TimeToStr(Time);
    sdCurrentPage: if D.ReportViewer <> nil then
                      D.Caption:= IntToStr(D.ReportViewer.CurrentPage)
                   else D.Caption:= '';
  End;
  DrawLabel(Band, D);
End;

procedure TTurboReportGenerator.AfterGenerate;
Begin
  GenerateBandforType(rbPageFooter);
End;

procedure TTurboReportGenerator.AfterGenerateData(Band: TTurboBand);
Begin
End;

procedure TTurboReportGenerator.BeforeGenerate;
Begin
  GenerateBandforType(rbPageHeader);
End;

procedure TTurboReportGenerator.BeforeGenerateData(Band: TTurboBand);
Begin
End;

procedure TTurboReportGenerator.newPage;
Begin
End;

Procedure TTurboReportGenerator.Control(Band: TTurboBand);
Begin
End;

Procedure TTurboReportGenerator.Calculate(Band: TTurboBand);
Begin
End;

Procedure TTurboReportGenerator.DrawLabel(Band: TTurboBand; L: TCustomLabel);
Begin
End;

Procedure TTurboReportGenerator.DrawImage(Band: TTurboBand; Img: TImage);
Begin
End;

Procedure TTurboReportGenerator.DrawDBImage(Band: TTurboBand; Img: TDBImage);
Begin
End;

Procedure TTurboReportGenerator.DrawShape(Band: TTurboBand; Sp: TShape);
Begin
End;

Procedure TTurboReportGenerator.DrawMemo(Band: TTurboBand; M: TCustomMemo);
Begin
End;

Procedure TTurboReportGenerator.DrawChart(Band: TTurboBand; M: TChart);
Begin
End;


Procedure TTurboReportGenerator.GenerateBandforType(BandType: TTurboBandType);
Var
  I: Integer;
Begin
    With fReport Do
    For I:= 0 to ControlCount - 1 Do
      if (Controls[I] is TTurboBand) and
         (TTurboBand(Controls[I]).BandType = BandType) Then
               GenerateBand(TTurboBand(Controls[I]));
End;

procedure TTurboReportGenerator.GenerateBand(Band: TTurboBand);
Begin
  if Assigned(Band.HeaderBand) Then GenerateBand(Band.HeaderBand);
  BeforeGenerateData(Band);
  if Assigned(Band.BeforeGenerate) then Band.BeforeGenerate(Band,  False);
  if Band.Height = 0 Then Exit;
  Control(Band);
  if (Band is TTurboGridBand) or (Band.DataSource = nil) Then
  Begin
    DrawFrame(Band);
    if Band is TTurboGridBand Then GenerateGridBand(TTurboGridBand(Band))
    else GenerateBandData(Band);
    Calculate(Band);
  End
  Else if ((Band.BandType <> rbSubDetail) and
           (Band.BandType <> rbSummary)) and
           (Band.BandType <> rbHeaderFooter) Then
  Begin
    if Assigned(Band.HeaderBand) Then GenerateBand(Band.HeaderBand);
    DrawFrame(Band);
    GenerateBandData(Band);
    Calculate(Band);
  End
  Else if ((Band.BandType = rbSubDetail) or
           (Band.BandType = rbSummary)) and
           (Band.DataSource <> nil) Then
  With Band.DataSource.DataSet Do
  Begin
    if not Active Then Open;
    First;
    While Not EOF Do
    BEgin
      BeforeGenerateData(Band);
      DrawFrame(Band);
      GenerateBandData(Band);
      Calculate(Band);
      Next;
      AfterGenerateData(Band);
    End;
  End;
  if Assigned(Band.FooterBand) Then GenerateBand(Band.FooterBand);
  if Assigned(Band.AfterGenerate) then Band.AfterGenerate(Band,  False);
  AfterGenerateData(Band);
End;

procedure TTurboReportGenerator.Generate(Report: TTurboReport);
  Procedure GenerateAllBand;
  Var
    I: Integer;
  Begin
    With Report Do
    For I:= 0 to ControlCount - 1 Do
      if (Controls[I] is TTurboBand) and
         (TTurboBand(Controls[I]).BandType = rbDetail) or
         (TTurboBand(Controls[I]).BandType = rbSubDetail) Then
               GenerateBand(TTurboBand(Controls[I]));
  End;

Begin
  freport:= Report;
  BeforeGenerate;
  NewPage;
  With Report Do
  if DataSource <> nil Then
    With DataSource.DataSet Do
    Begin
      if not Active Then Open;
      First;
      While Not EOF Do
      Begin
        GenerateAllBand;
        Next;
      End;
      First;
    End
  Else GenerateAllBand;
  GenerateBandforType(rbSummary);
  AfterGenerate;
End;

Constructor TTurboWMFGenerator.Create;
Begin
  Inherited;
  PageList:= TTurboPageList.Create;
End;

Destructor TTurboWMFGenerator.Destroy;
Begin
  PageList.Destroy;
  Inherited;
End;

Procedure TTurboWMFGenerator.DrawFrame(Band: TTurboBand);
Procedure DrawLine(X1, Y1, X2, Y2: Integer);
Begin
  With ReportCanvas Do
  Begin
    MoveTo(XStart + X1, YStart + Y1);
    LineTo(XStart + X2, YStart + Y2);
  End;
End;
Var
  B: TBrush;
  W: Integer;
Begin
  B:= TBrush.Create;
  B.Assign(ReportCanvas.Brush);
  ReportCanvas.Brush.Assign(Band.Brush);
  ReportCanvas.FillRect(Rect(XStart, YStart, XStart + Band.Width, YStart + Band.Height));
  With Band Do
  Begin
    ReportCanvas.Pen.Assign(Frame.Pen);
    W:= Frame.Pen.Width;
    if Frame.Left   Then DrawLine(0, 0, 0, Height - W);
    if Frame.Top    Then DrawLine(0, 0, Width - W, 0);
    if Frame.Bottom Then DrawLine(0, Height - W, Width - W, Height - W);
    if Frame.Right  Then DrawLine(Width - W, 0, Width - W, Height - W);
  End;
  ReportCanvas.Brush.Assign(B);
  B.Free;
End;

function  TTurboWMFGenerator.GetPageCount: Integer;
Begin
  Result:= PageList.Count;
End;

function TTurboWMFGenerator.GetPages(Indx: Cardinal): TPrintPage;
begin
  Result:= PageList.GetPage(Indx - 1);
end;

procedure TTurboWMFGenerator.AfterGenerate;
Begin
  ReportCanvas.Free;
End;

procedure TTurboWMFGenerator.BeforeGenerate;
Var
  I: Integer;
Begin
  PageList.Free;
  PageList:= TTurboPageList.Create;
  fCurrentPage:= 0;
  HFooter:= 0;
  For I:= 0 to fReport.ControlCount - 1 Do
    if (fReport.Controls[I] is TTurboBand) and
       (TTurboBand(fReport.Controls[I]).BandType = rbPageFooter) Then
             HFooter:= HFooter + TTurboBand(freport.Controls[I]).Height;
  HHeader:= 0;
  For I:= 0 to fReport.ControlCount - 1 Do
    if (fReport.Controls[I] is TTurboBand) and
       (TTurboBand(fReport.Controls[I]).BandType = rbPageHeader) Then
           HHeader:= HHeader + TTurboBand(freport.Controls[I]).Height;
End;

procedure TTurboWMFGenerator.BeforeGenerateData(Band: TTurboBand);
Begin
  ColumnControl(Band);
End;

Procedure TTurboWMFGenerator.Control(Band: TTurboBand);
Var
  X: Integer;
  S: String;
Begin
  X:= Band.Height;
  if Band.BandType = rbPageFooter Then X:= 0;
  S:= Band.Name;
  if (YStart + Band.Height > freport.Height - X -
          Trunc(freport.Margins.Bottom * PixelsInMM)) Then
  Begin
    CurrentCol:= CurrentCol + 1;
    if CurrentCol < fReport.Margins.Column Then
    Begin
      XStart:= XStart + Band.Width + Trunc(fReport.Margins.ColumnSpace * PixelsInMM);
      YStart:= Trunc(fReport.Margins.Top  * PixelsInMM) + HHeader;
    End
    Else Begin ReportCanvas.Free; NewPage; End;
  End;
End;

Procedure TTurboWMFGenerator.ColumnControl(Band: TTurboBand);
  Function BandWidth: Integer;
  Var
    W, I: Integer;
  Begin
    W:= 0;
    With fReport.Margins Do
    Begin
      I:= Trunc(((ColumnSpace * (Column - 1)) + Left + Right) * PixelsInMM);
      I:= (fReport.Width - I) div Column;
      W:= I;
    End;
    Result:= W;
  End;
Begin
  if (Band.bandType <> rbPageFooter) and
     (YStart > freport.Height - Band.Height - HFooter -
          Trunc(freport.Margins.Bottom * PixelsInMM)) Then
  Begin
    CurrentCol:= CurrentCol + 1;
    if CurrentCol < fReport.Margins.Column Then
    Begin
      XStart:= XStart + Band.Width + Trunc(fReport.Margins.ColumnSpace * PixelsInMM);
      YStart:= Trunc(fReport.Margins.Top  * PixelsInMM) + HHeader;
    End
    Else Begin ReportCanvas.Free; NewPage; End;
  End;
End;

Procedure TTurboWMFGenerator.Calculate(Band: TTurboBand);
Begin
  YStart:= YStart + Band.Height;
End;

procedure TTurboWMFGenerator.newPage;
Begin
  ReportPage:= TPrintPage.Create;
  ReportPage.Height:= fReport.Height;
  ReportPage.Width := fReport.Width;
  ReportCanvas:= TPrintCanvas.Create(ReportPage, 0);
  PageList.Add(ReportPage);
  ReportCanvas.Brush.Color:= clWhite;
  ReportCanvas.FillRect(rect(0, 0, fReport.Width, fReport.Height));
  CurrentCol:= 0;
  XStart    := 0;
  YStart    := 0;
  fCurrentPage:= PageList.Count;
  XStart:= Trunc(fReport.Margins.Left * PixelsInMM);
  YStart:= fReport.Height - HFooter -
          Trunc((fReport.Margins.Bottom + fReport.Margins.Top) * PixelsInMM);
  GenerateBandforType(rbPageFooter);
  XStart:= Trunc(fReport.Margins.Left * PixelsInMM);
  YStart:= Trunc(fReport.Margins.Top  * PixelsInMM);
  GenerateBandforType(rbPageHeader);
End;

Function TTurboWMFGenerator.RectControl(Band: TTurboBAnd; Var R: TRect; Var B: Boolean): TRect;
Var
  W: Integer;
Begin
  B:= False;
  W:= fReport.Width;
  With fReport Do
  if R.Right > W - Trunc((Margins.Left + Margins.Right) * PixelsInMM)
    Then R.Right:= W - Trunc((Margins.Left + Margins.Right) * PixelsInMM);
  w:= fReport.Height;
  With fReport Do
  if R.Bottom > W - Trunc((Margins.Top + Margins.Bottom) * PixelsInMM) Then
  Begin
    YStart:= Height + 1;
    ColumnControl(Band);
    B:= True;
  End;
  Result:= R;
end;

Procedure TTurboWMFGenerator.DrawLabel(Band: TTurboBand; L: TCustomLabel);
Var
  R: TRect;
  T: Integer;
  S1: String;
  A: TAlignment;
  B: Boolean;
Begin
  if L is TLabel Then
  Begin
    A:= TLabel(L).Alignment;
    S1:= TLabel(L).Caption;
    ReportCanvas.Brush.Color:= TLabel(L).Color;
    ReportCanvas.Font.Assign(TLabel(L).Font);
  End
  else if L is TDBText Then
  Begin
//    A:= TDBTExt(L).Alignment;
    A:= TDBText(L).Field.Alignment;
//    if TDBText(L).Field.DisplayText <> '' Then
//       S1:= TDBText(L).Field.DisplayText
      //FmtStr(S1, TDBText(L).Field.DisplayFormat, [TDBText(L).Field.curValue])
//    else
    S1:= TDBText(L).Field.DisplayText;
    ReportCanvas.Brush.Color:= TDBText(L).Color;
    ReportCanvas.Font.Assign(TDBText(L).Font);
  End
  else if L is TTurboSysData Then
  Begin
    A:= TTurboSysData(L).Alignment;
    S1:= TTurboSysData(L).Caption;
    ReportCanvas.Brush.Color:= TTurboSysData(L).Color;
    ReportCanvas.Font.Assign(TTurboSysData(L).Font);
  End;
  T:= DT_LEFT;
  case A of
    taCenter      : T:= DT_CENTER;
    taRightJustify: T:= DT_RIGHT;
  End;
  T:= T or DT_EXPANDTABS;
  R:= Rect(XStart + L.Left, YStart + L.Top, XStart + L.Left + L.Width, YStart + L.Top + L.Height);
  ReportCanvas.FillRect(R);
  DrawText(ReportCanvas.Handle, PChar(S1), System.Length(S1), R, T);
End;

Procedure TTurboWMFGenerator.DrawChart(Band: TTurboBand; M: TChart);
Var
  R: Trect;
  WMF: TMetaFile;
Begin
  WMF:= M.TeeCreateMetafile(False, Rect(0, 0, M.Width, M.Height));
  R:= Rect(XStart + Band.Left + M.Left,
           YStart + M.Top,
           XStart + M.Left + M.Width,
           YStart + M.Top + M.Height);
  if CurrentCol = 0 Then
  Begin
    R.Left := R.Left - XStart;
    R.Right:= R.Right - XStart;
  End;
  With fReport Do
  if R.Right > Width -
             Trunc((Margins.Left + Margins.Right) * PixelsInMM) Then
  Begin
    R.Right:= Width - Trunc((Margins.Left + Margins.Right) * PixelsInMM);
    ReportCanvas.StretchDraw(R, WMF);
  End
  Else
    ReportCanvas.Draw(R.Left, R.Top, WMF);
  WMF.Free;
End;

Procedure TTurboWMFGenerator.DrawImage(Band: TTurboBand; Img: TImage);
Var
  R: Trect;
Begin
  R:= Rect(XStart + Band.Left + Img.Left,
           YStart + Img.Top,
           XStart + Img.Left + Img.Width,
           YStart + Img.Top + Img.Height);
  if CurrentCol = 0 Then
  Begin
    R.Left := R.Left - XStart;
    R.Right:= R.Right - XStart;
  End;
  with fReport Do
    if R.Right > Width -
             Trunc((Margins.Left + Margins.Right) * PixelsInMM) Then
  Begin
    R.Right:=Width - Trunc((Margins.Left + Margins.Right) * PixelsInMM);
    ReportCanvas.StretchDraw(R, Img.Picture.Graphic);
  End
  Else if not Img.Stretch Then
    ReportCanvas.Draw(R.Left, R.Top, Img.Picture.Graphic)
  Else if Img.Stretch Then
    ReportCanvas.StretchDraw(R, Img.Picture.Graphic);
End;

Procedure TTurboWMFGenerator.DrawDBImage(Band: TTurboBand; Img: TDBImage);
Var
  R: Trect;
Begin
  R:= Rect(XStart + Band.Left + Img.Left,
           YStart + Img.Top,
           XStart + Img.Left + Img.Width,
           YStart + Img.Top + Img.Height);
  if CurrentCol = 0 Then
  Begin
    R.Left := R.Left - XStart;
    R.Right:= R.Right - XStart;
  End;
  With fReport Do
  if R.Right > Width -
             Trunc((Margins.Left + Margins.Right) * PixelsInMM) Then
  Begin
    R.Right:= Width - Trunc((Margins.Left + Margins.Right) * PixelsInMM);
    ReportCanvas.StretchDraw(R, Img.Picture.Graphic);
  End
  Else if not Img.Stretch Then
    ReportCanvas.Draw(R.Left, R.Top, Img.Picture.Graphic)
  Else if Img.Stretch Then
    ReportCanvas.StretchDraw(R, Img.Picture.Graphic);
End;

Procedure TTurboWMFGenerator.DrawMemo(Band: TTurboBand; M: TCustomMemo);
Var
  T, X, Y: Integer;
  R: TRect;
  S: String;
  B: Boolean;
  A: TAlignment;
Begin
  X:= XStart + M.Left;
  Y := YStart + M.Top;
  R:= Rect(X, Y, X + M.Width, Y + M.Height);
  RectControl(Band, R, B);
  if B Then
  Begin
    X:= XStart + M.Left;
    Y := YStart + M.Top;
    R:= Rect(X, Y, X + M.Width, Y + M.Height);
  End;
  if m is TMemo Then A:= TMemo(m).Alignment
  else if m is TDBMemo Then A:= TDBMemo(m).Alignment;
  Case A Of
    taLeftJustify : T:= DT_LEFT;
    taRightJustify: T:= DT_RIGHT;
    taCenter      : T:= DT_CENTER;
  End;
  T:= T or DT_EXPANDTABS or DT_WORDBREAK;
  if M is TDBMemo Then
  Begin
    ReportCanvas.Font.Assign(TDBMemo(M).Font);
    ReportCanvas.Brush.Color:= TDBMemo(M).Color;
  End
  else
  Begin
    ReportCanvas.Brush.Color:= TMemo(M).Color;
    ReportCanvas.Font.Assign(TMemo(M).Font);
  End;
  S:= M.Text;
  DrawText(ReportCanvas.Handle, PChar(S), System.Length(S), R, T);
End;

Procedure TTurboWMFGenerator.DrawShape(Band: TTurboBand; Sp: TShape);
Var
  X, Y, W, H, S: Integer;
Begin
  Sp.Visible:= False;
  X := XStart + Sp.Left;
  Y := YStart + Sp.Top;
  With ReportCanvas Do
  Begin
    Brush.Assign(Sp.Brush);
    Pen.Assign(Sp.Pen);
    X:= X +  Pen.Width div 2;
    W := Sp.Width - Pen.Width + 1;
    H := Sp.Height - Pen.Width + 1;
    if Pen.Width = 0 then
    begin
      Dec(W);
      Dec(H);
    end;
    if W + X - XStart > Band.Width  Then W:= Band.Width  - X;
    if H + Y - YStart > Band.Height Then H:= Band.Height - Y;    if W < H then S := W else S := H;
    if Sp.Shape in [stSquare, stRoundSquare, stCircle] then
    begin
      Inc(X, (W - S) div 2);
      Inc(Y, (H - S) div 2);
      W := S;
      H := S;
    end;
    case Sp.Shape of
      stRectangle, stSquare:
        Rectangle(X, Y, X + W, Y + H);
      stRoundRect, stRoundSquare:
        RoundRect(X, Y, X + W, Y + H, S div 4, S div 4);
      stCircle, stEllipse:
        Ellipse(X, Y, X + W, Y + H);
    end;
  End;
End;

Constructor TTurboReportView.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  ControlStyle := ControlStyle + [csAcceptsControls];
  align:= alClient;
  Color:= clWhite;
  fPaintBox:= nil;
  fZoom:= 100;
  fPageChange := nil;
  fAfterPrint := nil;
  fBeforePrint:= nil;
  fPrint      := nil;
  fZoomChange := nil;
  fReportGenerator:= nil;
  fReportDirectory:= 'C:\';
End;

Destructor TTurboReportView.Free;
Begin
  if fpaintBox <> nil Then fPaintBox.Free;
  if fReportGenerator <> nil Then fReportGenerator.Free;
  Inherited Free;
End;

Procedure TTurboReportView.SetCurrentPage(Value: Integer);
Begin
  if (Value > 0) and (Value <= fPageCount) Then
  Begin
     fCurrentPage:= Value;
     Invalidate;
  End;
End;

Procedure TTurboReportView.SetZoom(Value: Integer);
Begin
  if (fZoom <> Value) and
     (Value > 49) and (Value < 251) Then
  Begin
    fZoom:= Value;
    PaintPreview(Self);
    if fPaintBox <> nil then fPaintBox.Invalidate;
    if Assigned(OnZoomChange) then OnZoomChange(Self, fZoom);
  End;
End;

procedure TTurboReportView.PaintPreview(Sender: TObject);
Var
  R: TRect;
  W, H: Integer;
Begin
  if not (csDesigning in ComponentState) Then
  Begin
    W:= (fReportGenerator.Pages[fCurrentPage].Width * Zoom) div 100;
    H:= (fReportGenerator.Pages[fCurrentPage].Height * Zoom) div 100;
    R:= Rect((Width - W) div 2, (Height - H) div 2, W, H);

    if R.Left < 2 Then R.Left:= 2;
    if R.Top  < 2 Then R.Top := 2;

    fPaintBox.Height:= H + 15;
    fPaintBox.Width := W + 15;
    HorzScrollBar.Range:= fPaintBox.Width  + 30;
    VertScrollBar.Range:= fPaintBox.Height + 30;

    With fPaintBox.Canvas Do
    Begin
      Pen.Color:= clBlack;
      Pen.Width:= 2;
      Brush.Color:= clWhite;
      FillRect(rect(R.LEft - 2, R.Top - 2, R.Right + 2, R.Bottom + 2));
      Rectangle(R.LEft - 2, R.Top - 2, R.Right + 2, R.Bottom + 2);
      StretchDraw(R, fReportGenerator.Pages[fCurrentPage]);
      if Assigned(OnPrint) Then OnPrint(Self, fCurrentPage);
    End;
  End;
End;

procedure TTurboReportView.Notification(AComponent: TComponent;
      Operation: TOperation);
Begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) Then
         if (AComponent = Report) then Report := nil
    else if (AComponent = ProgressBar) then ProgressBar:= nil;
End;

Procedure TTurboReportView.First;
Begin
  if fCurrentPage <> 1 Then
  Begin
    CurrentPage:= 1;
    if Assigned(OnPageChange) Then OnPageChange(Self, fCurrentPage);
  End;
End;

Procedure TTurboReportView.Last;
Begin
  if fCurrentPage <> fPageCount Then
  Begin
    CurrentPage:= fPageCount;
    if Assigned(OnPageChange) Then OnPageChange(Self, fCurrentPage);
  End;
End;

Procedure TTurboReportView.Next;
Begin
  if fCurrentPage < fPageCount Then
  Begin
    fCUrrentPage:= fCurrentPage + 1;
    Invalidate;
    if Assigned(OnPageChange) Then OnPageChange(Self, fCurrentPage);
  End;
End;

Procedure TTurboReportView.Prev;
Begin
  if fCurrentPage > 1 Then
  Begin
    CUrrentPage:= fCurrentPage - 1;
    if Assigned(OnPageChange) Then OnPageChange(Self, fCurrentPage);
  End;
End;

Procedure TTurboReportView.GenerateExcel(L: TStrings);
Var
  T: TTurboExcelGenerator;
Begin
  Try
    T:= TTurboExcelGenerator.Create;
    T.Generate(Report);
    if L <> nil Then L.Assign(T.Lines);
    T.Exec;
  Finally
    T.Free;
  end;
End;

Procedure TTurboReportView.GenerateText(L: TStrings; fSave: Boolean);
Var
  T: TTurboTextGenerator;
  S: String;
Begin
  Try
    CharWInPixel:= fReport.Width div fReport.Printer.ColWidth;
    CharYInPixel:= fReport.Height div fReport.Printer.ColHeight;
    T:= TTurboTextGenerator.Create;
    T.Generate(Report);
    if L <> nil Then L.Assign(T.Lines);
    S:= freportDirectory;
    if S = '' Then S:= 'C:';
    if S[Length(S)] = '\' Then System.Delete(S, Length(S), 1);
    if ReportName <> '' Then
      S:= S + '\' + ReportName
    else S:= S + '\Turbo Report';
    if FSave Then T.Lines.SaveToFile(S);
  Finally
    T.Free;
  end;
End;

Procedure TTurboReportView.GenerateHTML(L: TStrings; fSave: Boolean);
Var
  T: TTurboHTMLGenerator;
  S: String;
Begin
  Try
    T:= TTurboHTMLGenerator.Create;
    S:= freportDirectory;
    if S[Length(S)] = '\' Then System.Delete(S, Length(S), 1);
    if ReportName <> '' Then
      T.FileName:= S + '\' + ReportName
    else T.FileName:= S + '\Turbo Report';
    T.PrinterTitle:= Report.Printer.Title;
    T.Generate(Report);
    if L <> nil Then L.Assign(T.Lines);
    if FSave Then T.Lines.SaveToFile(T.FileName + '.HTML');
  Finally
    T.Free;
  end;
End;

Procedure TTurboReportView.Generate;
Begin
  if Assigned (ProgressBar) Then ProgressBar.Visible:= False;
  if fReportGenerator <> nil Then fReportGenerator.Free;
  fReportGenerator:= TTurboWMFGenerator.Create;
  fReportGenerator.Generate(Report);
  HorzScrollBar.Tracking:=True;
  HorzScrollBar.Increment:=16;
  VertScrollBar.Tracking:=True;
  VertScrollBar.Increment:= 16;
  VertScrollBar.Position:= 0;
  HorzScrollBar.Position:= 0;

  if fPaintBox <> nil Then fPaintBox.Free;
  fPaintBox        := TPaintBox.Create(Self);
  fPaintBox.Parent := Self;
  fPaintBox.OnPaint:= PaintPreview;
  fPaintBox.Height := Height;
  fPaintBox.Width  := Width;
  fPaintBox.Left   := 5;
  fPaintBox.Top    := 5;
  Report.Visible   := False;
  fPageCount       := fReportGenerator.PageCount;
  CurrentPage      := 1;
  Color            := clBtnFace;
End;

constructor TTurboNavigator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] + [csOpaque];
  if not NewStyleControls then ControlStyle := ControlStyle + [csFramed];
  FVisibleButtons := [pgFirst, pgPrior, pgNext, pgLast, pgZoomIn, pgZoomOut,
        pgPrint, pgSetupPrinter, pgExcel, pgHTML, pgReFresh];
  FHints := TStringList.Create;
  TStringList(FHints).OnChange := HintsChanged;
  InitButtons;
  BevelOuter := bvNone;
  BevelInner := bvNone;
  Width := 275;
  Height := 25;
  ButtonWidth := 0;
  FocusedButton := pgFirst;
  FullRepaint := False;
  ShowHint:= True;
end;

destructor TTurboNavigator.Destroy;
begin
  FHints.Free;
  inherited Destroy;
end;

procedure TTurboNavigator.InitButtons;
var
  I: TPageCtlButton;
  Btn: TTurboBtn;
  X: Integer;
  ResName: string;
begin
  MinBtnSize := Point(20, 18);
  X := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    Btn := TTurboBtn.Create (Self);
    Btn.Flat := Flat;
    Btn.Index := I;
    Btn.Visible := I in FVisibleButtons;
    Btn.Enabled := True;
    Btn.SetBounds (X, 0, MinBtnSize.X, MinBtnSize.Y);
    FmtStr(ResName, 'PAGE%s', [BtnTypeName[I]]);
    Btn.Glyph.LoadFromResourceName(HInstance, ResName);
    Btn.NumGlyphs := 2;
    Btn.Enabled := False;
    Btn.Enabled := True;
    Btn.OnClick := ClickHandler;
    Btn.OnMouseDown := BtnMouseDown;
    Btn.Parent := Self;
    Buttons[I] := Btn;
    X := X + MinBtnSize.X;
  end;
  Buttons[pgFirst].Enabled:= False;
  Buttons[pgPrior].Enabled:= False;
  InitHints;
end;

procedure TTurboNavigator.InitHints;
var
  I: Integer;
  J: TPageCtlButton;
begin
  for J := Low(Buttons) to High(Buttons) do
    Buttons[J].Hint := BtnHintId[J];
  J := Low(Buttons);
  for I := 0 to (FHints.Count - 1) do
  begin
    if FHints.Strings[I] <> '' then Buttons[J].Hint := FHints.Strings[I];
    if J = High(Buttons) then Exit;
    Inc(J);
  end;
end;

procedure TTurboNavigator.HintsChanged(Sender: TObject);
begin
  InitHints;
end;

procedure TTurboNavigator.SetFlat(Value: Boolean);
var
  I: TPageCtlButton;
begin
  if FFlat <> Value then
  begin
    FFlat := Value;
    for I := Low(Buttons) to High(Buttons) do
      Buttons[I].Flat := Value;
  end;
end;

procedure TTurboNavigator.SetHints(Value: TStrings);
begin
  FHints.Assign(Value);
end;

procedure TTurboNavigator.SetVisible(Value: TPageButtonSet);
var
  I: TPageCtlButton;
  W, H: Integer;
begin
  W := Width;
  H := Height;
  FVisibleButtons := Value;
  for I := Low(Buttons) to High(Buttons) do
    Buttons[I].Visible := I in FVisibleButtons;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  Invalidate;
end;

procedure TTurboNavigator.AdjustSize (var W: Integer; var H: Integer);
var
  Count: Integer;
  MinH: Integer;
  MinW: Integer;
//  LastBtn,
  I: TPageCtlButton;
  Space, Temp, Remain: Integer;
  X, Y: Integer;
  ButtonHeight: Integer;
begin
  if (csLoading in ComponentState) then Exit;
  if Buttons[pgFirst] = nil then Exit;

  Count := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Inc(Count);
    end;
  end;
  if Count = 0 then Inc(Count);

  if ((FDirection = dirAuto) and (W >= H)) or (FDirection = dirHorizontal) then
  begin
    MinW := Count * MinBtnSize.X;
    if W < MinW then W := MinW;
    if H < MinBtnSize.Y then H := MinBtnSize.Y;

    ButtonWidth := W div Count;
    Temp := Count * ButtonWidth;
    if Align = alNone then W := Temp;

    X := 0;
    Remain := W - Temp;
    Temp := Count div 2;
    for I := Low(Buttons) to High(Buttons) do
    begin
      if Buttons[I].Visible then
      begin
        Space := 0;
        if Remain <> 0 then
        begin
          Dec(Temp, Remain);
          if Temp < 0 then
          begin
            Inc(Temp, Count);
            Space := 1;
          end;
        end;
        Buttons[I].SetBounds(X, 0, ButtonWidth + Space, Height);
        Inc(X, ButtonWidth + Space);
      end
      else
        Buttons[I].SetBounds (Width + 1, 0, ButtonWidth, Height);
    end;
    End
    else
    begin
      MinH := Count * (MinBtnSize.Y);
      if H < MinH then H := MinH;
      if W < MinBtnSize.X then W := MinBtnSize.X;
      ButtonHeight := (H div Count);
      Temp := Count * (ButtonHeight);
      if Align = alNone then H := Temp;
      Y := 0;
      Remain := H - Temp;
      Temp := Count div 2;
      for I := Low(Buttons) to High(Buttons) do
      begin
        if Buttons[I].Visible then
        begin
          Space := 0;
          if Remain <> 0 then
          begin
            Dec (Temp, Remain);
            if Temp < 0 then
            begin
              Inc (Temp, Count);
              Space := 1;
            end;
          end;
          Buttons[I].SetBounds (0, Y, Width, ButtonHeight + Space);
          Inc (Y, ButtonHeight + Space);
        end
        else
          Buttons[I].SetBounds (0, Height, ButtonHeight, Width);
        end;
      end;
end;

procedure TTurboNavigator.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  if not HandleAllocated then AdjustSize (W, H);
  inherited SetBounds (ALeft, ATop, W, H);
end;

procedure TTurboNavigator.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds(Left, Top, W, H);
  Message.Result := 0;
end;

procedure TTurboNavigator.ClickHandler(Sender: TObject);
begin
  BtnClick (TTurboBtn(Sender).Index);
end;

procedure TTurboNavigator.BtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  OldFocus: TPageCtlButton;
begin
  OldFocus := FocusedButton;
  FocusedButton := TTurboBtn(Sender).Index;
  if TabStop and (GetFocus <> Handle) and CanFocus then
  begin
    SetFocus;
    if (GetFocus <> Handle) then
      Exit;
  end
  else if TabStop and (GetFocus = Handle) and (OldFocus <> FocusedButton) then
  begin
    Buttons[OldFocus].Invalidate;
    Buttons[FocusedButton].Invalidate;
  end;
end;

procedure TTurboNavigator.BtnClick(Index: TPageCtlButton);
begin
  if not (csDesigning in ComponentState) and
         (Assigned(ReportViewer)) then
  with ReportViewer do
  begin
    case Index of
      pgPrior       : Prev;
      pgNext        : Next;
      pgFirst       : First;
      pgLast        : Last;
      pgZoomIn      : Zoom:= Zoom + 25;
      pgZoomOut     : Zoom:= Zoom - 25;
      pgPrint       : Print;
      pgSetupPrinter: SetupPrinter;
      pgExcel       : GenerateExcel(nil);
      pgHTML        : GenerateHTML(nil, True);
      pgReFresh     : Generate;
    end;
    Buttons[pgFirst].Enabled:= ReportViewer.CurrentPage > 1;
    Buttons[pgPrior].Enabled:= ReportViewer.CurrentPage > 1;
    Buttons[pgLast].Enabled:= ReportViewer.CurrentPage < ReportViewer.PageCount;
    Buttons[pgNext].Enabled:= ReportViewer.CurrentPage < ReportViewer.PageCount;
  end;
  if not (csDesigning in ComponentState) and Assigned(FOnBtnClick) then
    FOnBtnClick(Self, Index);
end;

procedure TTurboNavigator.WMSetFocus(var Message: TWMSetFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TTurboNavigator.WMKillFocus(var Message: TWMKillFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TTurboNavigator.KeyDown(var Key: Word; Shift: TShiftState);
var
  NewFocus: TPageCtlButton;
  OldFocus: TPageCtlButton;
begin
  OldFocus := FocusedButton;
  case Key of
    VK_RIGHT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus < High(Buttons) then
            NewFocus := Succ(NewFocus);
        until (NewFocus = High(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_LEFT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus > Low(Buttons) then
            NewFocus := Pred(NewFocus);
        until (NewFocus = Low(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_SPACE:
      begin
        if Buttons[FocusedButton].Enabled then
          Buttons[FocusedButton].Click;
      end;
  end;
end;

procedure TTurboNavigator.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TTurboNavigator.ActiveChanged;
var
  I: TPageCtlButton;
begin
  if not (Enabled) then
    for I := Low(Buttons) to High(Buttons) do
      Buttons[I].Enabled := False;
end;

procedure TTurboNavigator.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if not (csLoading in ComponentState) then
    ActiveChanged;
end;

procedure TTurboNavigator.Loaded;
var
  W, H: Integer;
begin
  inherited Loaded;
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  InitHints;
  ActiveChanged;
end;

procedure TTurboNavigator.FSetDirection(Value: TNavigatorDirection);
var
  W, H: Integer;
begin
  W := Width;
  H := Height;
  FDirection := value;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height)
    then inherited SetBounds (Left, Top, W, H);
  Invalidate;
end;

Function TTurboSysData.getCaption: String;
Begin
  case SysData of
    sdDate       : Result:= 'Date';
    sdDateTime   : Result:= 'Date, Time';
    sdTime       : Result:= 'Time';
    sdCurrentPage: Result:= 'CurrentPage';
  End;
End;

procedure TTurboSysData.Paint;
Begin
  Caption:= getCaption;
  Inherited;
End;

constructor TTurboSysData.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  AutoSize:= False;
  fSysData:= sdDate;
End;

Constructor TTurboTextGenerator.Create;
Begin
  Inherited;
  CurrentCol:= 0;
  HHeader:= 0;
  HFooter:= 0;
  fCurrentPage:= 0;
  XStart:= 0;
  YStart:= 0;
  PageList:= TStringList.Create;
end;

Destructor TTurboTextGenerator.Destroy;
Begin
  PageList.Free;
  Inherited;
End;

procedure TTurboTextGenerator.BeforeGenerate;
Var
  I: Integer;
Begin
  PageList.Free;
  PageList:= TStringList.Create;
  fCurrentPage:= 0;
  HFooter:= 0;
  For I:= 0 to fReport.ControlCount - 1 Do
    if (fReport.Controls[I] is TTurboBand) and
       (TTurboBand(fReport.Controls[I]).BandType = rbPageFooter) Then
             HFooter:= HFooter + Trunc(TTurboBand(fReport.Controls[I]).Height / CharYInPixel);
  HHeader:= 0;
  For I:= 0 to fReport.ControlCount - 1 Do
    if (fReport.Controls[I] is TTurboBand) and
       (TTurboBand(fReport.Controls[I]).BandType = rbPageHeader) Then
           HHeader:= HHeader + Trunc(TTurboBand(fReport.Controls[I]).Height / CharYInPixel);
End;

Function TTurboTextGenerator.getCurLine: Integer;
Begin
  if fCurrentPage = 0 Then Result:= YStart
  Else Result:= (fCurrentPage - 1) * fReport.Printer.ColHeight + YStart;
  if Result > PageList.Count - 1 Then Result:= PageList.Count - 1;
End;

procedure TTurboTextGenerator.newPage;
Var
  M: Array[0..BuffSize] Of Char;
  S: String;
  I: Integer;
Begin
  FillChar(M, SizeOf(M), ' ');
  M[fReport.Printer.ColWidth]:= #0;
  S:= StrPas(M);
  For I:= 0 to fReport.Printer.ColHeight Do PageList.Add(S);
  CurrentCol:= 0;
  XStart    := 0;
  YStart    := 0;
  fCurrentPage:= fCurrentPage + 1;
  XStart:= fReport.Margins.Left;
  YStart:= fReport.Printer.ColHeight - fReport.Margins.Top -
                  fReport.Margins.Bottom - HHeader + 1;
  GenerateBandforType(rbPageFooter);
  XStart:= fReport.Margins.Left;
  YStart:= fReport.Margins.Top;
  GenerateBandforType(rbPageHeader);
End;

Procedure TTurboTextGenerator.DrawMemo(Band: TTurboBand; M: TCustomMemo);
Var
  S, S1: String;
  A: TAlignment;

  Procedure mLength;
  Var
    X: Integer;
    F, SS: String;
    C: Array[0..BuffSize] of Char;
  Begin
    // X:= Round(M.Width / CharWInPixel);
    X:= fReport.Printer.ColWidth div fReport.Margins.Column;
    F:= '%-*s';
    if A = taRightJustify Then F:= '%*s';
    if A = taCenter Then X:= X Div 2;
    SS:= Format(F, [X, S1]);
    StrPCopy(C, SS);
    if A = taCenter Then C[X * 2]:= #0
    else C[X]:= #0;
    S1:= strPas(C);
  end;

Var
  XPos, Y, I: Integer;
Begin
  if m is TMemo Then A:= TMemo(m).Alignment
  else if m is TDBMemo Then A:= TDBMemo(m).Alignment;
  XPos:= XStart + Round(M.Left / CharWInPixel) + 1;
  Y:= getCurLine;
  For I:= 0 to M.Lines.Count - 1 do
  Begin
    if YStart >= fReport.Printer.ColHeight - HHeader - HFooter Then
      ColumnControl(Band);
    S1:= M.Lines[I];
    mLength;
    S:= PageList.Strings[Y + I];
    if Length(S) > XPos Then
    Begin
      System.Delete(S, XPos, Length(S1));
      System.Insert(S1, S, XPos);
    End
    Else S:= S + S1;
    PageList.Strings[Y + I]:= S;
    YStart:= YStart + 1;
  End;
  YStart:= YStart - Trunc(Band.Height / CharYInPixel) + 1;
End;

Procedure TTurboTextGenerator.ColumnControl(Band: TTurboBand);
Begin
  if (Band.bandType <> rbPageFooter) and (YStart >= fReport.Printer.ColHeight - HHeader - HFooter) Then
  Begin
    CurrentCol:= CurrentCol + 1;
    if CurrentCol < fReport.Margins.Column Then
    Begin
      XStart:= XStart + (fReport.Printer.ColWidth div freport.MArgins.Column) +
               Trunc(freport.MArgins.ColumnSpace / CharWInPixel);
      YStart:= HHeader;
    End
    Else NewPage;
  End
End;

Procedure TTurboTextGenerator.Calculate(Band: TTurboBand);
Begin
  YStart:= YStart + Trunc(Band.Height / CharYInPixel);
  ColumnControl(Band);
End;

Procedure TTurboTextGenerator.DrawLabel(Band: TTurboBand; L: TCustomLabel);
Var
  XPos: Integer;
  S1, S: String;
  A: TAlignment;

  Procedure mLength;
  Var
    X: Integer;
    F, SS: String;
    C: Array[0..BuffSize] of Char;
  Begin
    X:= Round(L.Width / CharWInPixel);
    F:= '%-*s';
    if A = taRightJustify Then F:= '%*s';
    if A = taCenter Then X:= X Div 2;
    SS:= Format(F, [X, S1]);
    StrPCopy(C, SS);
    if A = taCenter Then C[X * 2]:= #0
    else C[X]:= #0;
    S1:= strPas(C);
  end;

Begin
  XPos:= XStart + Round(L.Left / CharWInPixel) + 1;
  if L is TLabel Then
  Begin
    A:= TLabel(L).Alignment;
    S1:= TLabel(L).Caption;
  End
  else if L is TDBText Then
  Begin
    A:= TDBTExt(L).Alignment;
    S1:= TDBText(L).Field.DisplayText;
  End
  else if L is TTurboSysData Then
  Begin
    A:= TTurboSysData(L).Alignment;
    S1:= TTurboSysData(L).Caption;
  End
  else Exit;
  S:= PageList.Strings[getCurLine];
  mLength;
  if Length(S) > XPos Then
  Begin
    System.Delete(S, XPos, Length(S1));
    System.Insert(S1, S, XPos);
  End
  Else S:= S + S1;
  PageList.Strings[getCurLine]:= S;
End;

Constructor TTurboHTMLGenerator.Create;
Begin
  Inherited;
  PageList:= TStringList.Create;
  fPrnTitle:= '';
  fName:= '';
end;

Destructor TTurboHTMLGenerator.Destroy;
Begin
  PageList.Free;
  Inherited;
End;

procedure TTurboHTMLGenerator.BeforeGenerate;
Begin
  PageList.Free;
  PageList:= TStringList.Create;
  PageList.Add('<html>');
  PageList.Add('<head>');
  PageList.Add('<meta http-equiv="Content-Type"');
  PageList.Add('content="text/html; charset=windows-1254">');
  PageList.Add('<meta name="GENERATOR" content="Turbo Report Ver 1.0 by Cneyt ELBOL">');
  PageList.Add('<title>' + PrinterTitle + '</title>');
  PageList.Add('</head>');
  PageList.Add('<body>');
  PageList.Add('<body bgcolor="#FFFFFF">');
  PicCount:= 0;
  GenerateBandforType(rbPageHeader);
End;

procedure TTurboHTMLGenerator.AfterGenerate;
Begin
  GenerateBandforType(rbPageFooter);
  PageList.Add('</body>');
  PageList.Add('</html>');
End;

Function TTurboHTMLGenerator.getFontStr(F: TFont): String;
Begin
  Result:= '<font size="' + IntToStr(F.Size div 4) + '"';
  if F.Name <> 'MS Sans Serif' then Result:= Result +  'face="' + F.Name + '"';
  if F.Color <> clNone Then
     Result:= Result + ' color="' + ControlColor(F.Color) + '">';
  if fsItalic in F.Style then Result:= Result + '<em>';
  if fsBold   in F.Style then Result:= Result + '<strong>';
End;

Function TTurboHTMLGenerator.getFontEndStr(F: TFont): String;
Begin
  Result:= '';
  if fsBold   in F.Style then Result:= Result + '</strong>';
  if fsItalic in F.Style then Result:= Result + '</em>';
  Result:= Result + '</font>';
End;

Function TTurboHTMLGenerator.ControlColor(C: TColor): String;
var
  h: Longint;
Begin
  h:= colorToRGB(C);
  result:= Format('#%02X%02X%02X', [getRValue(h), getGValue(h), getBValue(h)]);
End;

Function TTurboHTMLGenerator.ControlProperties(C: TControl): String;
Var
  A: TAlignment;
  mC: TColor;
  Al, Cl, Capt: String;
  F: TFont;
Begin
  F:= TFont.Create;
  if C is TTurboSysData Then
  Begin
    A:= taLeftJustify;
    mC:= TTurboSysData(C).Color;
    Capt:= TTurboSysData(C).Caption;
    F.Assign(TTurboSysData(C).Font);
  end
  else if C is TLabel Then
  Begin
    A:= TLabel(C).Alignment;
    mC:= TLabel(C).Color;
    Capt:= TLabel(C).Caption;
    F.Assign(TLabel(C).Font);
  end
  else if C is TDBMemo Then
  Begin
    A:= TDBMemo(C).Alignment;
    mC:= TDBMemo(C).Color;
    Capt:= TDBMemo(C).Text;
    F.Assign(TDBMemo(C).Font);
  End
  else if C is TMemo Then
  Begin
    A:= TMemo(C).Alignment;
    mC:= TMemo(C).Color;
    Capt:= TMemo(C).Text;
    F.Assign(TMemo(C).Font);
  End
  else if C is TDBText Then
  Begin
    A:= TDBText(C).Field.Alignment;
    mC:= TDBText(C).Color;
    F.Assign(TDBText(C).Font);
    if (TDBText(C).Field <> nil) Then
    Begin
      Capt:= TDBText(C).Field.DisplayText;
      {
    case TDBText(C).Field.DataType Of
      ftString  : Capt:= TDBText(C).Field.AsString;
      ftInteger,
      ftWord    : Capt:= IntToStr(TDBText(C).Field.AsInteger);
      ftFloat   : Capt:= FloatToStr(TDBText(C).Field.AsFloat);
      ftCurrency: Capt:= Format('%M', [TDBText(C).Field.AsCurrency]);
      ftDate,
      ftDateTime: Capt:= DateToStr(TDBText(C).Field.AsDateTime);
    end
    }
  End
  End
  else
  Begin
    A:= taLeftJustify;
    mC:= clNone;
  End;
  case A of
    taLeftJustify : Al:= 'left';
    taCenter      : Al:= 'center';
    taRightJustify: Al:= 'right';
  End;
  Al:=  '  <td align="' + Al + '" width="' + IntToStr(C.Width + 20) + '"';
  Cl:= ControlColor(mC);
  if mC <> clNone Then Al:= Al + ' bgColor="' + Cl + '">'
  Else Al:= Al + '>';
  PageList.Add(Al);
  PageList.Add('  ' + getFontStr(F) + Capt + '  ' + getFontEndStr(F));
  F.Free;
End;

procedure TTurboHTMLGenerator.BeforeGenerateData(Band: TTurboBand);
Begin
  if (Band.BandType = rbHeaderFooter) or (Band.BandType = rbPageHeader) or
     (Band.BandType = rbPageFooter) Then
     PageList.Add('  <table border="3">')
  else
    PageList.Add('  <table border="0">');
  PageList.Add('  <tr>');
End;

procedure TTurboHTMLGenerator.AfterGenerateData(Band: TTurboBand);
Begin
  PageList.Add('  </tr>');
  PageList.Add('  </table>');
  PageList.Add('  ');
End;

Procedure TTurboHTMLGenerator.SaveImage(Img: TGraphic; ImgClass: TGraphicClass);
Var
  JPGImage: TJPEGImage;
  Nm: String;
  Bmp: TBitmap;
Begin
  if ImgClass = TMetaFile Then
  Begin
    Try
      Bmp:= TBitmap.Create;
      BMP.Height:= Img.Height;
      BMP.Width := Img.Width;
      Bmp.Canvas.Draw(0, 0, Img);
      Nm:= FNAme + IntToStr(PicCount) + '.JPG';
      Try
        JPGImage:= TJPEGImage.Create;
        JPGImage.Assign(Bmp);
        JPGImage.SaveToFile(Nm);
      Finally
        JPGImage.Free;
      End;
    Finally
      Bmp.Free;
    End;
  End
  Else
  Begin
    Nm:= FName + IntToStr(PicCount) + '.JPG';
    Try
      JPGImage:= TJPEGImage.Create;
      JPGImage.Assign(Img);
      JPGImage.SaveToFile(Nm);
    Finally
      JPGImage.Free;
    End;
  End;
  PageList.Add('  ');
  PageList.Add('  <td align="center" width="' + IntToStr(Img.Width) + '">');
  PageList.Add('  <img src="file:///' + Nm + '"width="' +
          IntToStr(Img.Width) + '"Height="' + IntToStr(Img.Height) + '" border="1">');
  PicCount:= PicCount + 1;
End;

Procedure TTurboHTMLGenerator.DrawLabel(Band: TTurboBand; L: TCustomLabel);
Begin
  ControlProperties(L);
End;

Procedure TTurboHTMLGenerator.DrawMemo(Band: TTurboBand; M: TCustomMemo);
Begin
  ControlProperties(M);
End;

Procedure TTurboHTMLGenerator.DrawImage(Band: TTurboBand; Img: TImage);
Begin
  if Img.Picture.ClassType = TMetaFile Then
      SaveImage(Img.Picture.Graphic, TMetaFile)
  else if Img.Picture.ClassType = TBitmap Then
      SaveImage(Img.Picture.Graphic, TBitmap)
  else if Img.Picture.ClassType = TJPEGImage Then
      SaveImage(Img.Picture.Graphic, TJPEGImage);
End;

Procedure TTurboHTMLGenerator.DrawDBImage(Band: TTurboBand; Img: TDBImage);
Begin
  SaveImage(Img.Picture.Graphic, TBitmap);
End;

Procedure TTurboHTMLGenerator.DrawChart(Band: TTurboBand; M: TChart);
Var
  WMF: TMetaFile;
Begin
  Try
    WMF:= M.TeeCreateMetafile(False, Rect(0, 0, M.Width, M.Height));
    SaveImage(WMF, TMetaFile)
  finally
    WMF.Free;
  End;
End;

constructor TTurboExcel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if not (csDesigning in ComponentState) then
  begin
    FDDE := TDdeClientConv.Create(nil);
    FDDE.ConnectMode := ddeManual;
    FDDE.OnOpen  := OpenLink;
    FDDE.OnClose := ShutDown;
  end;
  SetExeName('Excel');
  FDecimals := 2;
  FBatch := False;
  FMin := 200;
  FMax := 250;
end;

destructor TTurboExcel.Destroy;
begin
  if not (csDesigning in ComponentState) then FDDE.Free;
  if FLines <> nil then FLines.Free;
  if FCells <> nil then FCells.Free;
  inherited Destroy;
end;

procedure TTurboExcel.SetExeName(const Value: TFileName);
begin
  Disconnect;
  FExeName := ChangeFileExt(Value, '');
  if not (csDesigning in ComponentState) then
    FDDE.ServiceApplication := FExeName;
end;

procedure TTurboExcel.SetConnect(const Value: Boolean);
begin
  if FConnected = Value then Exit;
  if Value then Connect
           else Disconnect;
end;

procedure TTurboExcel.SetMin(const Value: Integer);
begin
  if Value > FMax then FMin := FMax
                  else FMin := Value;
end;

procedure TTurboExcel.SetMax(const Value: Integer);
begin
  if Value < FMin then FMax := FMin
                  else FMax := Value;
end;

function TTurboExcel.GetSelection: string;
begin
  Result := Request('Selection');
end;

function TTurboExcel.GetReady: Boolean;
begin
  Result := 'Ready' = Request('Status');
end;

procedure TTurboExcel.DoRect(Top, Left, Bottom, Right: Integer;
                        Data: TStrings; Request: Boolean);
  var
      i : Integer;
      Sel, Item : string;
      RowMark,
      ColMark : Char;
      Reply : PChar;
begin
  Wait;
  Select(1, 1);                          
  Sel := Selection;
  i := Pos('!', Sel);
  if i = 0 then raise Exception.Create(msgNoTable);
  RowMark := Sel[i+1];                   { Some nationalized version }
  ColMark := Sel[i+3];                   {  using other then R and C }
  FDDE.OnOpen  := nil;
  FDDE.OnClose := nil;                   { Disable event handlers }
try
  FDDE.SetLink('Excel', Copy(Sel, 1, i-1));  { Topic = Sheet name }
  ProcessMessages;
  if not FDDE.OpenLink then
    raise Exception.Create(msgNoLink);
  ProcessMessages;
  Item := Format('%s%d%s%d:%s%d%s%d', [RowMark, Top, ColMark, Left,
                                    RowMark, Bottom, ColMark, Right]);
  if Request then
  begin
    Reply := FDDE.RequestData(Item);
    if Reply <> nil then Data.SetText(Reply);
    StrDispose(Reply);
  end
  else if not FDDE.PokeDataLines(Item, Data) then
    raise Exception.Create('"'+ Item + msgNotAccepted);
finally
  ProcessMessages;
  LinkSystem;
  ProcessMessages;
  FDDE.OpenLink;
  FDDE.OnOpen  := OpenLink;              { Enable event handlers }
  FDDE.OnClose := ShutDown;
  if not Connected and Assigned(FOnClose) then FOnClose(Self);
end; end;

procedure TTurboExcel.LinkSystem;
begin
  FDDE.SetLink('Excel', 'System');
end;

procedure TTurboExcel.CheckConnection;
begin
  if not Connected then
    raise Exception.Create(msgNoConnect);
end;

procedure TTurboExcel.OpenLink(Sender: TObject);
begin
  FConnected := True;
  if Assigned(FOnOpen) then FOnOpen(Self);
end;

procedure TTurboExcel.ShutDown(Sender: TObject);
begin
  FConnected := False;
  if Assigned(FOnClose) then FOnClose(Self);
end;

procedure TTurboExcel.LocateExcel;
  const
      BuffSize = 255;
  var
      Buff: array[0..BuffSize] of Char;
      Fn  : string;
      Len : Longint;
begin
  Len := BuffSize;
  StrPCopy(Buff, '.XLS');
  if (RegQueryValue(HKEY_CLASSES_ROOT, Buff, Buff, Len)
    = ERROR_SUCCESS) and (StrScan(Buff,'E') <> nil) then
  begin
    StrCat(Buff, '\Shell\Open\Command');
    Len := BuffSize;
    if RegQueryValue(HKEY_CLASSES_ROOT, Buff, Buff, Len)
      = ERROR_SUCCESS then
    begin
      Fn := StrPas(StrUpper(Buff));
      Len := Pos('EXCEL.EXE', Fn);
      Delete(Fn, Len + Length('EXCEL.EXE'), 255);
      if Buff[0] = '"' then Delete(Fn, 1, 1);
      if FileExists(Fn) then
        ExeName := Fn;
    end;
  end;
end;

procedure TTurboExcel.Connect;
begin
  if FConnected then Exit;
  LinkSystem;
  if FDDE.OpenLink then Exit;
  LocateExcel;
  if FDDE.OpenLink then Exit;            { Try again }
  ProcessMessages;
  if FDDE.OpenLink then Exit;            { Once more }
  raise Exception.Create(msgNoExcel);
end;

procedure TTurboExcel.Disconnect;
begin
  if FConnected then FDDE.CloseLink;
end;

procedure TTurboExcel.Wait;
  const
        TryCount = 64;
  var
      i : Integer;
begin
  i := 0;
  repeat
    if Ready then Break;    { Waiting for Excel }
    Inc(i);
  until i = TryCount;
  if i = TryCount then
    raise Exception.Create(msgNoRespond);
end;

procedure TTurboExcel.ProcessMessages;
begin
{$IFDEF WIN32}
  Application.HandleMessage;
{$ELSE}
  Application.ProcessMessages;
{$ENDIF}
end;

function TTurboExcel.Request(const Item: string): string;
  var
      Reply : PChar;
begin
  CheckConnection;
  ProcessMessages;
  Reply := FDDE.RequestData(Item);
  if Reply = nil then Result := msgNoReply
                 else Result := StrPas(Reply);
  StrDispose(Reply);
end;

procedure TTurboExcel.Exec(const Cmd: string);
  var
      a : array[0..555] of Char;
begin
  CheckConnection;
  StrPCopy(a, Cmd);
  if FDDE.ExecuteMacro(a, False) then
    ProcessMessages
  else
  begin
    Wait;
    if FDDE.ExecuteMacro(a, True ) then
      ProcessMessages
    else
      raise Exception.Create('"' + Cmd + msgNotAccepted);
  end
end;

procedure TTurboExcel.Run(const Mn: string);
begin
  if FMacro = '' then
    raise Exception.Create(msgNoMacro);
  Exec('[RUN("' + FMacro + '!' + Mn + '";FALSE)]');
end;

procedure TTurboExcel.Select(Row, Col: Integer);
begin
  Exec(Format('[SELECT("R%dC%d")]', [Row, Col]));
end;

procedure TTurboExcel.PutStr(Row, Col: Integer; const s: string);
  procedure SendMin;
    var
        i : Integer;
  begin
    FCells.Clear;
    for i:=0 to FMin-1 do
    begin
      FCells.Add(FLines[0]);          { FCells as work space }
      FLines.Delete(0);
    end;
    DoRect(FFirstRow, FFirstCol, FFirstRow + FMin - 1, FLastCol,
           FCells, False);
    Inc(FFirstRow, FMin);
  end;

  procedure DoBatch;
    var
        i, j, Index : Integer;
        Line : string;
  begin
    Index := Row - FFirstRow;                { Index to modify }
    if Index >= Lines.Count then
      for i:=Lines.Count to Index do         { Expand if needed }
        Lines.Add('');
    if Lines.Count > FMax then               { Send if needed }
    begin
      SendMin;
      Index := Row - FFirstRow;              { Recalc Index }
    end;
    if Col > FLastCol then FLastCol := Col;  { Adjust to max }
    Line := Lines[Index];
    FCells.Clear;                            { Empty FCells }
    j := 1;
    for i:=1 to Length(Line) do              { Line disasseble }
      if Line[i] = #9 then begin
                             FCells.Add(Copy(Line, j, i-j));
                             j := i + 1;
                           end;
    FCells.Add(Copy(Line, j, Length(Line) + 1 - j));
    if FCells.Count < Col - FFirstCol + 1 then
      for i:=FCells.Count to Col-FFirstCol do{ Expand if needed }
        FCells.Add('');
    FCells[Col-FFirstCol] := s;              { Replace cell }
    Line := FCells[0];
    for i:=1 to FCells.Count-1 do            { Line reasseble }
      Line := Line + #9 + FCells[i];
    Lines[Index] := Line;                    { Replace line }
  end;

begin           { TTurboExcel.PutStr }
  if BatchOn and (Col >= FFirstCol) and (Row >= FFirstRow) then
    DoBatch
  else
    Exec(Format('[FORMULA("%s","R%dC%d")]', [s, Row, Col]));
end;

procedure TTurboExcel.PutExt(Row, Col: Integer; e: Extended);
begin
  PutStr(Row, Col, Format('%0.*f', [Decimals, e]));
end;

procedure TTurboExcel.PutMoney(Row, Col: Integer; e: Double);
begin
  PutStr(Row, Col, Format('%m', [e]));
end;

procedure TTurboExcel.PutInt(Row, Col: Integer; i: Longint);
begin
  PutStr(Row, Col, IntToStr(i));
end;

procedure TTurboExcel.PutDay(Row, Col: Integer; d: TDateTime);
begin
  PutStr(Row, Col, DateToStr(d));
end;

procedure TTurboExcel.BatchStart(FirstRow, FirstCol: Integer);
begin
  if FLines = nil then FLines := TStringList.Create
                  else FLines.Clear;
  if FCells = nil then FCells := TStringList.Create
                  else FCells.Clear;
  FFirstRow := FirstRow;
  FFirstCol := FirstCol;
  FLastCol  := FirstCol;
  FBatch := True;
end;

procedure TTurboExcel.BatchCancel;
begin
  if FLines <> nil then FLines.Free;
  if FCells <> nil then FCells.Free;
  FLines := nil;
  FCells := nil;
  FBatch := False;
end;

procedure TTurboExcel.BatchSend;
begin
  if FLines <> nil then
    DoRect(FFirstRow, FFirstCol, FFirstRow + FLines.Count - 1,
           FLastCol, FLines, False);
  BatchCancel
end;

procedure TTurboExcel.GetRange(R: TRect; Lines: TStrings);
begin
  DoRect(R.Top, R.Left, R.Bottom, R.Right, Lines, True);
end;

function TTurboExcel.GetCell(Row, Col: Integer): string;
  var
      Data : TStringList;
begin
  Result := msgNoReply;
  Data := TStringList.Create;
try
  DoRect(Row, Col, Row, Col, Data, True);
  if Data.Count = 1 then Result := Data[0];
finally
  Data.Free
end; end;

procedure TTurboExcel.OpenMacroFile(const Fn: TFileName; Hide: Boolean);
begin
  if FMacroPath = Fn then Exit;
  CloseMacroFile;
  Exec('[OPEN("' + Fn + '")]');
  if Hide then  Exec('[HIDE()]');
  FMacroPath := Fn;
  FMacro := ExtractFileName(Fn);
end;

procedure TTurboExcel.CloseMacroFile;
begin
  if FMacro <> '' then
  try
    Exec('[UNHIDE("' + FMacro + '")]');
    Exec('[ACTIVATE("' + FMacro + '")]');
    Exec('[CLOSE(FALSE)]');
  finally
    FMacro := '';
    FMacroPath := '';
  end;
end;

Constructor TTurboExcelGenerator.Create;
Begin
  Inherited;
  XStart:= 1;
  YStart:= 1;
  PageList:= TStringList.Create;
  PageList.Add('[NEW(1)]');
End;

Procedure TTurboExcelGenerator.Exec;
Var
  E: TTurboExcel;
  I: Integer;
Begin
  Try
    E:= TTurboExcel.Create(Application);
    E.Connect;
    For I:= 0 To PageList.Count - 1 Do
       E.Exec(PageList.Strings[I]);
  finally
    E.Free;
  End;
End;

Destructor TTurboExcelGenerator.Destroy;
Begin
  PageList.Free;
  Inherited;
End;

Procedure TTurboExcelGenerator.DrawLabel(Band: TTurboBand; L: TCustomLabel);
Var
  S: String;
Begin
  if L is TTurboSysData Then
  Begin
    S:= TTurboSysData(L).Caption;
  end
  else if L is TLabel Then
  Begin
    S:= TLabel(L).Caption;
  end
  else if L is TDBText Then
  Begin
    if (TDBText(L).Field <> nil) Then
    Begin
    case TDBText(L).Field.DataType Of
      ftString  : S:= TDBText(L).Field.AsString;
      ftInteger,
      ftWord    : S:= IntToStr(TDBText(L).Field.AsInteger);
      ftFloat   : S:= FloatToStr(TDBText(L).Field.AsFloat);
      ftCurrency: S:= Format('%M', [TDBText(L).Field.AsCurrency]);
      ftDate,
      ftDateTime: S:= DateToStr(TDBText(L).Field.AsDateTime);
    end
  End;
  End;
  PageList.Add(Format('[FORMULA("%s","R%dC%d")]', [s, YStart, XStart]));
  XStart:= XStart + 1;
End;

Procedure TTurboExcelGenerator.DrawMemo(Band: TTurboBand; M: TCustomMemo);
  function DeleteChars(S: String): string;
  Var
    I: Integer;
  Begin
    I:= System.Pos('"', S);
    While I > 0 do
    Begin
      System.Delete(S, I, 1);
      I:= System.Pos('"', S);
    End;
    Result:= S;
  End;

  Procedure AddLines(L: TStrings);
  Var
    I: Integer;
    S: String;
  Begin
    XStart:= XStart + 1;
    For I:= 0 to L.Count - 1 Do
    Begin
      S:= L.Strings[I];
      S:= DeleteChars(S);
      if S = '' then S:= '     ';
      PageList.Add(Format('[FORMULA("%s","R%dC%d")]', [S, YStart, XStart]));
      YStart:= YStart + 1;
    End;
  End;

Begin
  if M is TDBMemo Then
  Begin
    AddLines(TDBMemo(M).Lines);
  End
  else if M is TMemo Then
  Begin
    AddLines(TMemo(M).Lines);
  End;
End;

procedure TTurboExcelGenerator.BeforeGenerateData(Band: TTurboBand);
Begin
  YStart:= YStart + 1;
  XStart:= 1;
End;

Procedure TTurboExcelGenerator.Calculate(Band: TTurboBand);
Begin
End;

Function TTurboExcelGenerator.getCurLine: Integer;
Begin
  Result:= 0;
End;

procedure TTurboReportView.SetupPrinter;
Var
  D: TPrinterSetupDialog;
begin
  Try
    D:= TPrinterSetupDialog.Create(Application);
    D.Execute;
  finally
    D.Free;
  End;
end;

procedure TTurboReportView.Print;
Var
  Start, Stop: Integer;

  Procedure PrintNormal;
  Var
    I: Integer;
  Begin
    For I:= Start + 1 to Stop + 1 Do
    Begin
      Printer.Canvas.StretchDraw(Rect(0, 0,
             Printer.PageWidth, Printer.PageHeight), fReportGenerator.Pages[I]);
      if Assigned (ProgressBar) Then ProgressBar.Position:=ProgressBar.Position+1;
      if Assigned(OnPrint) Then OnPrint(Self, I);
      if I < PageCount Then Printer.NewPage;
    End;
  End;

  Procedure PrintFast;
  Var
    E, H, I: Integer;
    F: TextFile;
    L: TstringList;
  Begin
    L:= TStringList.Create;
    GenerateText(L, False);
    H:= Report.Printer.ColHeight;
    E:= (Stop + 1) * H;
    if E > L.Count Then E:= L.Count;
    AssignFile(F, Report.Printer.Port); //'C:\Gecici\Deneme.Txt');
    ReWrite(F);
    For I:= Start * H + Start to E do // Downto 0 Do
    Begin
      WriteLn(F, L.Strings[I]);
      if Assigned (ProgressBar) Then ProgressBar.Position:=ProgressBar.Position+1;
      if Assigned(OnPrint) Then OnPrint(Self, I);
    End;
    CloseFile(F);
    if Assigned(OnPrint) Then OnPrint(Self, 0);
  End;

Var
  D: TPrintDialog;
begin
  Try
    if Assigned(OnBeforePrint) Then OnBeforePrint(Self, 0);
    D:= TPrintDialog.Create(Application);
    D.Options:= [poPrintToFile, poPageNums, poSelection, poWarning];
    D.FromPage := 1;
    D.MinPage := 1;
    D.ToPage := PageCount;
    D.MaxPage := PageCount;
    if D.Execute Then
    Begin
      with D do
      begin
        if PrintRange = prAllPages then
        begin
          Start := MinPage - 1;
          Stop := MaxPage - 1;
          end
        else if PrintRange = prSelection then
        begin
          Start := CurrentPage;
          Stop := Start;
        end
        else  { PrintRange = prPageNums }
        begin
          Start := FromPage - 1;
          Stop := ToPage - 1;
        end;
      End;
      Printer.Title:= Report.Printer.Title;
      Printer.BeginDoc;
      if Assigned (ProgressBar) Then
      Begin
        ProgressBar.Visible:= True;
        ProgressBar.Position:= 0;
      End;
      if Report.Printer.Mode = prNormal Then PrintNormal
      else if Report.Printer.Mode = prFast Then PrintFast;
      Printer.EndDoc;
    End;
  finally
    D.Free;
    if Assigned(OnAfterPrint) Then OnAfterPrint(Self, PageCount);
  End;
end;

{ TTurboGridBand }

procedure TTurboGridBand.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = Grid) then Grid:= nil
end;

procedure TTurboReportGenerator.GenerateGridBand(Band: TTurboGridBand);
  Procedure DrawShape1(X, W: Integer);
  Var
    S: TShape;
  Begin
      try
      S:= TShape.Create(Band);
      S.Shape:= stRectangle;
      S.Width:= W;
      S.Left:= X;
      S.Top:= 0;
      S.Height:= Band.Height;
      S.Pen.Width:= 1;
      DrawShape(Band, S);
    finally
      S.Free;
    End;
  End;

  Procedure AddLabel(mCaption: String; X, W, H: Integer; A: TAlignment);
  Var
    L: TLabel;
  Begin
    Try
      L:= TLabel.Create(Band);
      if Band.Grid is TStringGrid Then
      Begin
        L.Font.Assign(TStringGrid(Band.Grid).Font);
        L.Color:= TStringGrid(Band.Grid).Color;
      End
      else if Band.Grid is TDBGrid Then
      Begin
        L.Font.Assign(TDBGrid(Band.Grid).Font);
        L.Color:= TDBGrid(Band.Grid).Color;
      End;
      L.Autosize:= False;
      L.Alignment:= A;
      L.Left := X + 3;
      L.Top  := 1;
      L.Width:= W - 1;
      L.Height:= H - 2;
      L.Caption:= mCaption;
      Band.Height:= H;
      DrawShape1(X, W + 3);
      DrawLabel(Band, L);
    Finally
      L.Free;
    End;
  End;

  Procedure GenerateString;
  Var
    X, mRow, mCol: Integer;
    S: String;
  Begin
    For mRow:= 0 to TStringGrid(Band.Grid).RowCount Do
    Begin
      X:= 1;
      BeforeGenerateData(Band);
      For mCol:= 0 to TStringGrid(Band.Grid).ColCount Do
      With TStringGrid(Band.Grid) Do
      Begin
        S:= IntToStr(mrow) + '\' + IntToStr(mCol);
        AddLabel(Cells[mCol, mRoW], X, ColWidths[mCol], RowHeights[mRow] + 3, taLeftJustify);
        X:= X + ColWidths[mCol] + 2;
      End;
      AfterGenerateData(Band);
      Calculate(Band);
    End;
  End;

  Procedure GenerateData;
  Var
    X, I: Integer;
    S: String;
    T: TTextMetric;
   Begin
    getTextMetrics(TDBGrid(Band.Grid).Canvas.Handle, T);
    if TDBGrid(Band.Grid).DataSource.Dataset <> nil Then
    With TDBGrid(Band.Grid).DataSource.Dataset Do
    Begin
      if not Active Then Open;
      X:= 1;
      BeforeGenerateData(Band);
      //DrawShape1(X - 1);
      With TDBGrid(Band.Grid) Do
        For I:= 0 to Columns.Count - 1 Do
          Begin
            S:= Columns.Items[I].Field.DisplayName;
            AddLabel(S, X, Columns.Items[I].Width, t.tmHeight + 2, taCenter);
            X:= X + Columns.Items[I].Width + 2;
          end;
      AfterGenerateData(Band);
      Calculate(Band);
      First;
      While not EoF Do
      Begin
        BeforeGenerateData(Band);
        X:= 1;
        //DrawShape1(X - 1);
        With TDBGrid(Band.Grid) Do
          For I:= 0 to Columns.Count - 1 Do
          Begin
              S:= Columns.Items[I].Field.DisplayText;
              AddLabel(S, X, Columns.Items[I].Width, t.tmHeight + 2,
                            Columns.Items[I].Field.Alignment);
              X:= X + Columns.Items[I].Width + 2;
            End;
        AfterGenerateData(Band);
        Calculate(Band);
        Next;
      End;
    End;
  End;

begin
  if Band.Grid is TStringGrid Then GenerateString
  else if Band.Grid is TDBGrid Then GenerateData;
end;

Procedure AddFieldFromGrid(Band: TTurboBand; Grid: TDBGrid);
  Procedure AddLabel(X, W, H: Integer; Fld: String);
  Var
    L: TDBText;
  Begin
    Try
      L:= TDBText.Create(Band);
      L.Parent:= Band;
      L.Font.Assign(Grid.Font);
      L.Color:= Grid.Color;
      L.Autosize:= False;
      L.Left := X;
      L.Top  := 1;
      L.Width:= W;
      L.Height:= H;
      L.DataSource:= Grid.DataSource;
      L.DataField := Fld;
      Finally
    End;
  End;

  Procedure AddDBGrid;
  Var
    X, I: Integer;
    T: TTextMetric;
    S: String;
   Begin
    getTextMetrics(Grid.Canvas.Handle, T);
    Band.Height:= t.tmHeight + 2;
    X:= 1;
    For I:= 0 To Grid.Columns.Count - 1 Do
    Begin
      AddLabel(X, Grid.Columns.Items[I].Width, t.tmHeight + 2, Grid.Columns[I].Field.FieldName);
      X:= X + Grid.Columns.Items[I].Width + 2;
    End;
  End;

Begin
  if Grid.DataSource.Dataset <> nil Then AddDBGrid;
End;

initialization
  PixelsInInch:= Screen.PixelsPerInch / 12.7;
  PixelsInMM  := PixelsInInch / 2;
  CharWInPixel:= 8;
  CharYInPixel:= 20;
  RegisterClasses([TTurboBand, TTurboBtn, TTurboExcel,
      TTurboFrame, TTurboGridBand, TTurboMargins, TTurboNavigator,
      TTurboPrinter, TTurboReport, TTurboReportView,
      TTurboSysData, TLabel, TDBText, TMemo,
      TDBMemo, TImage, TDBImage, TTable, TQuery, TDataSource]);
finalization
  ;
end.
